SystemOrganization addCategory: #'Aggregate-BTree'!
SystemOrganization addCategory: #'Aggregate-Core'!
SystemOrganization addCategory: #'Aggregate-Variables'!
SystemOrganization addCategory: #'Aggregate-Model'!
SystemOrganization addCategory: #'Aggregate-Indices'!
SystemOrganization addCategory: #'Aggregate-Errors'!
SystemOrganization addCategory: #'Aggregate-Test'!
SystemOrganization addCategory: #'Aggregate-Introspection'!
SystemOrganization addCategory: #'Aggregate-Example'!

Notification subclass: #ARDynamicVariable
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Variables'!

ARDynamicVariable subclass: #ARCurrentClient
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Variables'!

ARDynamicVariable subclass: #ARCurrentContainer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Variables'!

!ARDynamicVariable class methodsFor: 'private' stamp: 'sebastian.sastre 11/19/2009 13:10'!
defaultValue
	^ nil! !

!ARDynamicVariable class methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 13:12'!
use: anObject during: aBlock
	"Make this dynamic variable to have anObject 
	as its value during the evaluation of aBlock."
	
	^ aBlock on: self do: [ :notification | 
		notification resume: anObject ]! !

!ARDynamicVariable class methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 13:10'!
value
	"Answer the dynamic value of the receiver."
	
	^ self signal! !

Notification subclass: #ARResetingWithOnlineClients
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Errors'!

BTreeKeysArray variableSubclass: #ARBTreeKeysArray
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-BTree'!

!ARBTreeKeysArray methodsFor: 'as yet unclassified' stamp: 'sebastian.sastre 11/21/2009 13:40'!
findIndexForKey: aMagnitude
	self withIndexDo:
		[:key :i |
		(key isNil or: [key > aMagnitude]) ifTrue:
			[^ i - 1]].
	^ self size
! !

BTree subclass: #ARBTree
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-BTree'!

!ARBTree class methodsFor: 'as yet unclassified' stamp: 'sebastian.sastre 11/21/2009 13:06'!
order: aNumber
	^ self keys: (ARBTreeKeysArray new: aNumber)! !

!ARBTree methodsFor: 'actions' stamp: 'sebastian.sastre 11/20/2009 12:36'!
at: key ifAbsent: absentReaction ifPresent: presentReaction
	"Lookup the given key in the receiver. 
	If it is present, answer the value of evaluating 
	presentReaction block with the value associated with the key. 
	Otherwise, evaluate absentReaction block."

	^ presentReaction value: (self at: key ifAbsent: absentReaction)
! !

!ARBTree methodsFor: 'actions' stamp: 'sebastian.sastre 11/21/2009 13:40'!
at: aMagnitude put: anObject
	"Stores anObject using aMagnitude to locate it.
	If the root was changed (some insertions need that), 
	the tree will mark itself as dirty"
	| oldRoot |

	oldRoot := root.
	super at: aMagnitude put: anObject.

	oldRoot = root ifFalse: [self markDirty].
	^ anObject! !

!ARBTree methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 18:27'!
fixKeys
	"Insertions and deletions use to produce splits in 
	the tree nodes that needs fixing"
	
	root fixKeys! !

!ARBTree methodsFor: 'initializing' stamp: 'sebastian.sastre 11/19/2009 20:12'!
initializeWithKeys: aBTreeKeys

	aBTreeKeys size > 3 ifFalse: [self error: 'The BTree order must be at least 4'].
	root := ARBTreeLeafNode keys: aBTreeKeys.
	self makePersistent! !

!ARBTree methodsFor: 'actions' stamp: 'sebastian.sastre 11/20/2009 17:28'!
remove: aMagnitude ifAbsent: aBlock
	"Removes anObject stored at aMagnitude 
	or evaluates aBlock if not found.
	If the root was changed in the remotion, 
	the tree will mark itself as dirty"
	
	| oldRoot |

	oldRoot := root.
	super removeKey: aMagnitude ifAbsent: aBlock.
	oldRoot = root ifFalse: [self markDirty]! !

!ARBTree methodsFor: 'actions' stamp: 'sebastian.sastre 11/20/2009 17:27'!
removeKey: aMagnitude
	"Removes anObject stored at aMagnitude.
	If the root was changed in the remotion, 
	the tree will mark itself as dirty"
	
	| oldRoot |
	oldRoot := root.
	super removeKey: aMagnitude.
	oldRoot = root ifFalse: [self markDirty]! !

BTreeInteriorNode variableSubclass: #ARBTreeInteriorNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-BTree'!

!ARBTreeInteriorNode class methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 18:53'!
keys: anArray
	"Answers a new instance of the receiver that 
	is made persistent."
	^ (super keys: anArray)
		makePersistent;
		yourself	! !

!ARBTreeInteriorNode methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 18:26'!
at: aMagnitude put: anObject
	"Stores anObject using aMagnitude to locate it.
	The tree's change demads a mark dirty to be written"

	self markDirty.
	^ super at: aMagnitude put: anObject! !

!ARBTreeInteriorNode methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 18:26'!
childForKey: aMagnitude
	| index |
	index := keys findIndexForKey: aMagnitude.
	index = 0 ifTrue:
		[self markDirty.
		keys at: 1 put: aMagnitude.
		^ self at: 1].
	^ self at: index
		
! !

!ARBTreeInteriorNode methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 18:27'!
fixKeys
	"Insertions and deletions use to produce splits in 
	the tree nodes that needs fixing"
	| key |
	key := (self at: 1) fixKeys.
	self firstKey = key ifFalse:[
		"Transcript show: '*'."
		keys at: 1 put: key.
		self markDirty].
	^ key! !

!ARBTreeInteriorNode methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 20:12'!
interiorNodeClass

	^ ARBTreeInteriorNode ! !

!ARBTreeInteriorNode methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 18:30'!
parent: aNode
	"Changes the parent of this node.
	It requires a mark dirty to be persisted."
	
	super parent: aNode.
	self markDirty! !

!ARBTreeInteriorNode methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 18:33'!
updateKey: oldMagnitude to: newMagnitude
	"Updates the node (or some parent's)
	regarding to old a new magnitudes.
	It requires a mark dirty to be persisted."

	self markDirty.
	^ super updateKey: oldMagnitude to: newMagnitude! !

TreeSet subclass: #ARTreeSet
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-BTree'!

!ARTreeSet class methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 18:45'!
defaultEqualitySelector
	"Answers the default selector to use for the equality check."
	
	"#isIdenticalTo: is about identity. It's used by OmniBase as #== but for persistent objects."
	^ #isIdenticalTo:! !

!ARTreeSet class methodsFor: 'as yet unclassified' stamp: 'sebastian.sastre 11/21/2009 13:08'!
defaultKeys
	^ ARBTreeKeysArray new: 64! !

!ARTreeSet class methodsFor: 'accessing' stamp: 'sebastian.sastre 11/21/2009 13:05'!
order: anInteger

	^ self basicNew 
		initializeWithKeys: (ARBTreeKeysArray new: anInteger) 
		sortSelector: self defaultSortSelector
		equalitySelector: self defaultEqualitySelector 
	! !

!ARTreeSet methodsFor: 'public' stamp: 'sebastiansastre 5/3/2012 11:18'!
add: anObject

	super add: anObject.
	
	self markDirty.
	
	^ anObject ! !

!ARTreeSet methodsFor: 'public' stamp: 'sebastian.sastre 3/7/2011 20:25'!
addAll: aCollection
	
	super addAll: aCollection.
	
	self markDirty.
	
	^ aCollection ! !

!ARTreeSet methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 18:38'!
fixKeys
	"Insertions and deletions use to produce splits in 
	the tree nodes that needs fixing"
	tree fixKeys! !

!ARTreeSet methodsFor: 'public' stamp: 'sebastiansastre 5/2/2012 15:52'!
ifAbsentAdd: anObject

	(self includes: anObject) ifTrue:[^ nil].
	
	super add: anObject.
	
	self markDirty.
	
	^ anObject ! !

!ARTreeSet methodsFor: 'initializing' stamp: 'sebastian.sastre 11/19/2009 20:13'!
initializeWithKeys: aBtreeKeys sortSelector: sortSymbol equalitySelector: equalsSymbol

	self makePersistent.
	self setTree: (ARBTree keys: aBtreeKeys) sortSelector: sortSymbol equalitySelector: equalsSymbol.
	self markDirty.! !

!ARTreeSet methodsFor: 'public' stamp: 'sebastian.sastre 3/7/2011 20:25'!
remove: anObject
	
	super remove: anObject.
	
	self markDirty.
	
	^ anObject ! !

!ARTreeSet methodsFor: 'public' stamp: 'sebastian.sastre 3/7/2011 20:26'!
removeAllSuchThat: aBlock
	
	super removeAllSuchThat: aBlock.
	
	self markDirty! !

!ARTreeSet methodsFor: 'accessing' stamp: 'sebastiansastre 10/13/2011 10:08'!
species

	^ ARTreeSet ! !

!TreeSet methodsFor: '*aggregate' stamp: 'sebastian.sastre 12/4/2009 11:59'!
copyFrom: start to: end
	| results |
	results := self species sortSelector: sortKey equalitySelector: equalsKey.
	start to: (self size min: end) do:[:i|
		results add: (self atIndex: i)].

	^ results! !

!TreeSet methodsFor: '*aggregate' stamp: 'sebastian.sastre 12/4/2009 12:01'!
first: anInteger
	"Answer a copy of the receiver comprising the leftmost anInteger elements."

	^self copyFrom: 1 to: anInteger! !

!TreeSet methodsFor: '*aggregate' stamp: 'sebastian.sastre 12/4/2009 12:01'!
last

	^ (self last: 1) first! !

!TreeSet methodsFor: '*aggregate' stamp: 'sebastian.sastre 12/4/2009 12:01'!
last: anInteger
	"Answer a copy of the receiver comprising the rightmost anInteger elements."

	| size |
	size := self size.
	^self copyFrom: (size + 1 - anInteger) to: size! !

!TreeSet methodsFor: '*aggregate' stamp: 'sebastian.sastre 5/13/2010 22:25'!
removeLast 

	self remove: self last! !

!TreeSet methodsFor: '*aggregate' stamp: 'sebastian.sastre 3/9/2011 22:14'!
select: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Collect into a new collection like the receiver, only those elements for 
	which aBlock evaluates to true. Answer the new collection."

	| newCollection |
	newCollection := self species new.
	newCollection sortKey: sortKey.
	self do: [:each | (aBlock value: each) ifTrue: [newCollection add: each]].
	^newCollection! !

!TreeSet methodsFor: '*aggregate' stamp: 'sebastian.sastre 3/9/2011 22:15'!
sortKey: aSelector

	sortKey := aSelector ! !

!OmniBase class methodsFor: '*aggregate' stamp: 'sebastiansastre 8/18/2011 15:34'!
nilKey
	"Answers a value where nil is mapped on a given indexing support"
	^ 'yw32ttedw207tmxqzhwhv7s3-nil'! !

Error subclass: #ARError
	instanceVariableNames: 'subject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Errors'!

ARError subclass: #ARCantIndexNil
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Errors'!

ARError subclass: #ARCantLockOdb
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Errors'!

ARError subclass: #ARCantStoreMany
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Errors'!

!ARCantStoreMany commentStamp: 'sebastian.sastre 11/26/2009 11:03' prior: 0!
This error is raised when something that is not many things was sent to be indexed.
In plain english this means: can't index with this an object that is not implementing #do:!

ARError subclass: #ARCantStoreNil
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Errors'!

!ARError class methodsFor: 'actions' stamp: 'sebastiansastre 6/5/2011 11:18'!
for: anObject 

	^ self new 
		subject: anObject;
		yourself   ! !

!ARError class methodsFor: 'actions' stamp: 'sebastiansastre 6/5/2011 11:18'!
signalFor: anObject 

	^ self new 
		subject: anObject;
		signal   ! !

!ARError methodsFor: 'accessing' stamp: 'sebastiansastre 6/5/2011 11:18'!
subject
	^ subject! !

!ARError methodsFor: 'accessing' stamp: 'sebastiansastre 6/5/2011 11:18'!
subject: anObject
	subject := anObject! !

ARError subclass: #ARIndexCollisionWithNil
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Errors'!

ARError subclass: #ARKeyAlreadyStored
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Errors'!

ARError subclass: #ARNoIndexForAggregate
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Errors'!

ARError subclass: #ARNoIndexForAttribute
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Errors'!

ARError subclass: #ARNoIndicesForAggregate
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Errors'!

BTreeLeafNode variableSubclass: #ARBTreeLeafNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-BTree'!

!ARBTreeLeafNode class methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 18:52'!
keys: anArray
	"Answers a new instance of the receiver that 
	is made persistent."
	^ (super keys: anArray)
		makePersistent;
		yourself	! !

!ARBTreeLeafNode methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 18:35'!
at: aMagnitude put: anObject
	"Stores anObject using aMagnitude to locate it.
	The tree's change demads a mark dirty to be written"

	self markDirty.
	^ super at: aMagnitude put: anObject! !

!ARBTreeLeafNode methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 18:36'!
fixKeys
	"Insertions and deletions use to produce splits in 
	the tree nodes that needs fixing.
	Interior nodes fix expects leaf to return the first key on fixKeys"
	^ self firstKey! !

!ARBTreeLeafNode methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 20:12'!
interiorNodeClass

	^ ARBTreeInteriorNode ! !

!ARBTreeLeafNode methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 18:37'!
parent: aNode
	"Changes the parent of this node.
	It requires a mark dirty to be persisted."
	
	super parent: aNode.
	self markDirty! !

!BlockContext methodsFor: '*aggregate' stamp: 'sebastian.sastre 11/19/2009 13:04'!
commitIn: anOmniBaseTransaction
	"Evaluate the receiver block while setting the current transaction 
	to anOmniBaseTransaction. If an error occurs during execution of
	this block the transaction will be rolled back. If there is no error the
	transaction will be committed. Answer the result of evaluating the block."

	| activeProcess previousTxn result |
	activeProcess := Processor activeProcess.
	previousTxn := OmniBase getCurrentAndSet: anOmniBaseTransaction for: activeProcess.
	result := [|answer|
				answer := self value.
				anOmniBaseTransaction commit.
				answer]
		ensure: [
			previousTxn isNil
				ifTrue: [OmniBase removeFor: activeProcess]
				ifFalse: [OmniBase setTransaction: previousTxn for: activeProcess]].
	^result! !

!BlockContext methodsFor: '*aggregate' stamp: 'sebastian.sastre 11/19/2009 22:01'!
evaluateIn: anOmniBaseTransaction
	"Evaluate the receiver block while setting the current transaction 
	to anOmniBaseTransaction. If an error occurs during execution of
	this block the transaction will be rolled back. Answer the result of
	the block."

	| activeProcess previousTxn |
	activeProcess := Processor activeProcess.
	previousTxn := OmniBase getCurrentAndSet: anOmniBaseTransaction for: activeProcess.
	^[self ifCurtailed: [anOmniBaseTransaction abort]]
		ensure: [
			previousTxn isNil
				ifTrue: [OmniBase removeFor: activeProcess]
				ifFalse: [OmniBase setTransaction: previousTxn for: activeProcess]]! !

!Collection class methodsFor: '*aggregate' stamp: 'sebastiansastre 9/25/2011 02:28'!
persistentReaderSupport

	^ ARCollectionPersistentReader! !

!UndefinedObject methodsFor: '*aggregate' stamp: 'sebastiansastre 8/18/2011 15:35'!
asBtreeKeyOfSize: keySize

	^ OmniBase nilKey asBtreeKeyOfSize: keySize! !

!BlockClosure methodsFor: '*aggregate' stamp: 'sebastian.sastre 7/20/2009 17:56'!
commitIn: anOmniBaseTransaction
		"Evaluate the receiver block while setting the current transaction 
		to anOmniBaseTransaction. If an error occurs during execution of
		this block the transaction will be rolled back. If there is no error the
		transaction will be committed. Answer the result of evaluating the block."

	| activeProcess previousTxn result |
	activeProcess := Processor activeProcess.
	previousTxn := OmniBase getCurrentAndSet: anOmniBaseTransaction for: activeProcess.
	[result := [[|answer|
				answer := self value.
				anOmniBaseTransaction commit.
				answer] 
					on: WAResponseNotification
					do:[:notification| 
						anOmniBaseTransaction commit.
						notification signal]] ensure:[
							anOmniBaseTransaction abort]]
		ensure: [
			previousTxn isNil
				ifTrue: [OmniBase removeFor: activeProcess]
				ifFalse: [OmniBase setTransaction: previousTxn for: activeProcess]].
	^result! !

TestCase subclass: #ARAggregateTestCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Test'!

ARAggregateTestCase subclass: #ARIndexingTest
	instanceVariableNames: 'pool users tags'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Test'!

ARIndexingTest subclass: #ARBTreeTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Test'!

!ARBTreeTest methodsFor: 'accessing' stamp: 'sebastian.sastre 3/9/2011 21:26'!
someAlphanumerics
	
	^ (1 to: 1000) collect:[:i|
		UUID new asString36]! !

!ARBTreeTest methodsFor: 'accessing' stamp: 'sebastian.sastre 3/9/2011 21:26'!
someNumbers

	| rnd |

	rnd := Random new.
	rnd seed: TimeStamp now asSeconds.
	
	^ (1 to: 1000) collect:[:i|
		rnd next]! !

!ARBTreeTest methodsFor: 'testing' stamp: 'sebastian.sastre 3/9/2011 21:40'!
testAlphanumeric
	
	| sorted tree elements |
	
	sorted := SortedCollection sortBlock: [:a :b| a < b].
	tree := BTree order: 8.
		
	elements := self someAlphanumerics.

	sorted addAll: elements.
	sorted do:[:e|
		tree at: (sorted indexOf: e) put: e].

	1 to: elements size do:[:i|
		self assert: (sorted at: i) = (tree at: i)].
	
	tree 
		from: 1 
		to: elements size 
		keysAndValuesDo: [:k :v|
			self assert: (sorted at: k) = v]
		

		
! !

!ARBTreeTest methodsFor: 'testing' stamp: 'sebastiansastre 5/3/2012 12:10'!
testInRoot
	
	pool commit:[
		self should: [pool root at: #place] raise: KeyNotFound].
	
	pool commit:[
		pool root at: #place put: (ARTreeSet sortSelector: #name equalitySelector: #=).
		self shouldnt: [pool root at: #place] raise: KeyNotFound].

	pool commit:[
		self shouldnt: [pool root at: #place] raise: KeyNotFound.
		self assert: (pool root at: #place) odbObjectID notNil.
		self assert: (pool root at: #place) odbGetOID notNil].

	pool commit:[
		pool root removeKey: #place.
		self should: [pool root at: #place] raise: KeyNotFound].
! !

!ARBTreeTest methodsFor: 'testing' stamp: 'sebastian.sastre 3/9/2011 21:34'!
testNumbers
	
	| sorted tree elements |
	

	
	sorted := SortedCollection sortBlock: [:a :b| a < b].
	tree := BTree order: 8.
		
	elements := self someNumbers.

	sorted addAll: elements.
	sorted do:[:e|
		tree at: (sorted indexOf: e) put: e].

	1 to: elements size do:[:i|
		self assert: (sorted at: i) = (tree at: i)]
		
! !

!ARBTreeTest methodsFor: 'testing' stamp: 'sebastian.sastre 3/9/2011 21:34'!
testNumbersSmall
	
	| sorted tree elements |
	
	
	
	sorted := SortedCollection sortBlock: [:a :b| a < b].
	tree := BTree order: 8.
		
	elements := #(
		0.000648575369570672
		0.000731434207750221
		0.001271925867196138
		0.001606590580943316
		0.003727566452570058
		0.00424666097585422
		).

	sorted addAll: elements.
	sorted do:[:e|
		tree at: (sorted indexOf: e) put: e].

	1 to: elements size do:[:i|
		self assert: (sorted at: i) = (tree at: i)]
		
! !

!ARIndexingTest methodsFor: 'actions' stamp: 'sebastian.sastre 11/26/2009 16:47'!
addDummyTags: anInteger times: anotherInteger

	tags := Set new.

	self addDummyTags: anInteger times: anotherInteger to: pool and: tags
	
! !

!ARIndexingTest methodsFor: 'actions' stamp: 'sebastian.sastre 11/26/2009 11:47'!
addDummyTags: anInteger times: anotherInteger to: aPool and: aSet

	1 to: anotherInteger do:[:i|
		aPool commit:[
			1 to: anInteger do:[:j| |tag|
				tag := ARDummyTag new.
				aSet add: tag.
				tag save]]]! !

!ARIndexingTest methodsFor: 'actions' stamp: 'sebastian.sastre 11/26/2009 16:47'!
addDummyUsers: anInteger times: anotherInteger

	users := Set new.

	self addDummyUsers: anInteger times: anotherInteger to: pool and: users
	
! !

!ARIndexingTest methodsFor: 'actions' stamp: 'sebastian.sastre 5/12/2010 14:31'!
addDummyUsers: anInteger times: anotherInteger do: aBlock

	users := Set new.

	self addDummyUsers: anInteger times: anotherInteger to: pool and: users do: aBlock 
	
! !

!ARIndexingTest methodsFor: 'actions' stamp: 'sebastian.sastre 5/12/2010 14:30'!
addDummyUsers: anInteger times: anotherInteger to: aPool and: aSet

	self addDummyUsers: anInteger times: anotherInteger to: aPool and: aSet do: [:e|]! !

!ARIndexingTest methodsFor: 'actions' stamp: 'sebastian.sastre 12/23/2010 11:26'!
addDummyUsers: anInteger times: anotherInteger to: aPool and: aSet do: aBlock

	anotherInteger timesRepeat:[
		aPool commit:[
			anInteger timesRepeat: [ |guy|
				guy := ARDummyUser new.
				aSet add: guy.
				aBlock value: guy.
				guy save]]]! !

!ARIndexingTest methodsFor: 'actions' stamp: 'sebastian.sastre 12/23/2010 09:22'!
benchmarksBasic
	| creation update deletion destruction |

	self setUp.

	creation := [1 to: 100 do:[:i|
		pool commit:[
			1to: 10 do:[:j|
				ARDummyPerson new save]]]] timeToRun.
	Transcript cr; show: 'creating 1000 aggregates in: '; show: creation asString; show: ' [msec]'.

	update := [|stream|
	stream := (pool readOnly:[ARDummyPerson findAll]) asArray readStream.
	[stream atEnd] whileFalse:[
		pool commit:[
			(stream next: 10) do:[:each|
				each fresh save]]]] timeToRun.
	Transcript cr; show: 'updating 1000 aggregates in: '; show: update asString; show: ' [msec]'.

	deletion := [|stream|
	stream := (pool readOnly:[ARDummyPerson findAll]) asArray readStream.
	[stream atEnd] whileFalse:[
		pool commit:[
			(stream next: 10) do:[:each|
				each fresh delete]]]] timeToRun.
	Transcript cr; show: 'deleting 1000 aggregates in: '; show: deletion asString; show: ' [msec]'.

	destruction := [|stream|
	stream := (pool readOnly:[ARDummyPerson findAll]) asArray readStream.
	[stream atEnd] whileFalse:[
		pool commit:[
			(stream next: 10) do:[:each|
				each fresh destroy]]]] timeToRun.
	Transcript cr; show: 'destroying 1000 aggregates in: '; show: destruction asString; show: ' [msec]'.
	
	self tearDown.
! !

!ARIndexingTest methodsFor: 'accessing' stamp: 'sebastian.sastre 12/23/2010 09:01'!
pool 

	^ pool ! !

!ARIndexingTest methodsFor: 'private' stamp: 'sebastian.sastre 11/30/2009 09:22'!
repositoryPath

	^ FileDirectory default fullName,FileDirectory slash,'AggregateTest'! !

!ARIndexingTest methodsFor: 'actions' stamp: 'sebastian.sastre 11/24/2009 12:47'!
resetStorage

	self resetStorageOf: pool! !

!ARIndexingTest methodsFor: 'actions' stamp: 'sebastian.sastre 4/26/2010 18:15'!
resetStorageOf: aPool

	aPool commit:[
		aPool root at: #storage put: ARStorage new.
		ARDummyUser indices do:[:anIndex| (aPool root at: #storage) addIndex: anIndex].
		ARDummyPerson indices do:[:anIndex| (aPool root at: #storage) addIndex: anIndex].
		ARDummyTag indices do:[:anIndex| (aPool root at: #storage) addIndex: anIndex]]! !

!ARIndexingTest methodsFor: 'private' stamp: 'sebastian.sastre 11/30/2009 09:22'!
secondaryRepositoryPath

	^ FileDirectory default fullName,FileDirectory slash,'AggregateSecondaryTest'! !

!ARIndexingTest methodsFor: 'initializing' stamp: 'sebastian.sastre 11/22/2009 15:36'!
setUp
	
	pool := ARODBPool onPath: self repositoryPath.
	
	pool exists ifFalse:[
		pool ensureExistence.
		self resetStorage]! !

!ARIndexingTest methodsFor: 'initializing' stamp: 'sebastian.sastre 11/20/2009 16:34'!
tearDown
	
	users := nil.
	
	pool isNil ifFalse:[|p|
		p := pool.
		pool := nil.
		p reset].
! !

!ARIndexingTest methodsFor: 'testing' stamp: 'sebastian.sastre 11/26/2009 14:15'!
testBasicUpdate

	| someId oldUsername newUsername |

	self resetStorage.
	self addDummyUsers: 10 times: 1.
	someId := users anyOne id.

	oldUsername := 'smeagol'.
	pool commit:[|guy|
		guy := (pool root at: #storage) find: ARDummyUser where: #id is: someId.
		self assert: guy notNil.
		self assert: (guy isIdenticalTo: (ARDummyUser atId: someId)).
		(ARDummyUser atId: someId) username: oldUsername; save].
	newUsername := 'gollum'.
	pool commit:[|user|
		user := ARDummyUser atUsername: oldUsername.
		self assert: user id = someId.
		user username: newUsername; save].
	
	pool readOnly:[
		self assert: (ARDummyUser atId: someId) username = newUsername.
		self assert: (ARDummyUser atId: someId) = (ARDummyUser atUsername: newUsername).
		self assert: (ARDummyUser atUsername: oldUsername) isNil].

! !

!ARIndexingTest methodsFor: 'testing' stamp: 'sebastian.sastre 11/24/2009 13:25'!
testCommitNestedToADifferentRepository
	"Test if ActiveRecords is able to make a commit inside
	of a commit made to a different repository"
	| moreUsers anotherPool guy1Id guy2Id girlId |

	self resetStorage.
	"put 10 active records in the test storage"
	self addDummyUsers: 10 times: 1.
	
	moreUsers := Set new.
	"put 10 active records more in a secondary test storage"
	anotherPool := ARODBPool onPath: self secondaryRepositoryPath.
	anotherPool exists ifFalse:[
		anotherPool ensureExistence].
	self resetStorageOf: anotherPool.
	self addDummyUsers: 10 times: 1 to: anotherPool and: moreUsers.

	pool commit:[|guys|
		guys := users copy.
		guy1Id := (guys remove: users anyOne) id.
		(ARDummyUser atId: guy1Id) username: 'frodo'; save.
		anotherPool commit:[|girls|
			girls := moreUsers copy.
			girlId := (girls remove: moreUsers anyOne) id.
			(ARDummyUser atId: girlId) username: 'galadriel'; save].
		guy2Id := (guys remove: guys anyOne) id.
		(ARDummyUser atId: guy2Id) username: 'golum'; save].
	
	anotherPool readOnly:[
		self assert: (ARDummyUser atId: girlId) username = 'galadriel'].

	pool readOnly:[
		anotherPool readOnly:[
			self assert: (ARDummyUser atId: girlId) username = 'galadriel'].
		self assert: (ARDummyUser atId: guy1Id) username = 'frodo'.
		self assert: (ARDummyUser atId: guy2Id) username = 'golum'].
	
! !

!ARIndexingTest methodsFor: 'testing' stamp: 'sebastiansastre 9/20/2011 17:35'!
testComposed1

	| user person good bad |

	self resetStorage.

	user := ARDummyUser new.
	person := ARDummyPerson new.
	user person: person.

	pool commit:[|personObjectID userObjectID|
		person save.
		personObjectID := pool currentTransaction getObjectIDOrNil: person.		
		user save.
		userObjectID := pool currentTransaction getObjectIDOrNil: user.
		self assert: (person isIdenticalTo: user person)].
	
	pool readOnly:[|guy|
		guy := person fresh.
		self assert: (guy isIdenticalTo: person fresh).
		self assert: (guy isIdenticalTo: user fresh person).
		self assert: (guy isIdenticalTo: user person fresh).
		self assert: (person isIdenticalTo: person aggregate).
		self assert: (guy isIdenticalTo: person fresh aggregate).

		"Pay attention to these ones..."
		self assert: (guy fresh = guy).
		self assert: (guy = guy fresh).
		self assert: (guy fresh isIdenticalTo: guy).
		self assert: (guy isIdenticalTo: guy).
		self assert: (guy isIdenticalTo: guy fresh).

		self assert: (guy fresh = guy fresh).
		self assert: (guy fresh isIdenticalTo: guy fresh)]! !

!ARIndexingTest methodsFor: 'testing' stamp: 'sebastiansastre 9/20/2011 17:24'!
testComposed2

	| user person good bad |

	self resetStorage.

	user := ARDummyUser new.
	person := ARDummyPerson new.
	user person: person.

	pool commit:[|personObjectID userObjectID|
		person save.
		personObjectID := pool currentTransaction getObjectIDOrNil: person.		
		user save.
		userObjectID := pool currentTransaction getObjectIDOrNil: user.
		self assert: (person isIdenticalTo: user person)].

	good := ARDummyTag new label: 'good'.
	bad := ARDummyTag new label: 'bad'.
	person tags add: good.

	pool commit:[
		person := person fresh.
		good save.
		bad save.
		person tags add: good.
		person save].	

	pool readOnly:[
		self assert: (person fresh tags anyOne isIdenticalTo: good fresh)].

	pool commit:[|g|
		g := good fresh.
		g label: 'terrific'; save.
		self assert: (person fresh tags anyOne label = 'terrific')].
	
	pool readOnly:[
		self assert: (person tags anyOne fresh label = 'terrific').
		self assert: (person fresh tags anyOne label = 'terrific').
		self assert: (person tags anyOne fresh isIdenticalTo: person fresh tags anyOne)].

	pool commit:[|g|
		good := good fresh.
		good save.
		self assert: (person fresh tags anyOne label = 'terrific')].

	pool readOnly:[
		self assert: (person tags anyOne fresh label = 'terrific')].
	
	pool readOnly:[|guy|
		guy := person fresh.
		self assert: (guy fresh isIdenticalTo: guy).
		self assert: (guy fresh isIdenticalTo: guy fresh)]! !

!ARIndexingTest methodsFor: 'testing' stamp: 'seb 9/3/2011 23:47'!
testDontAllowNil

	| oldId newId |

	self resetStorage.
	self addDummyUsers: 10 times: 1.
	oldId := users anyOne id.


	self 
		should:[pool commit:[(ARDummyUser atId: oldId) username: nil; save]] 
		raise: ARCantIndexNil.
	
! !

!ARIndexingTest methodsFor: 'testing' stamp: 'sebastian.sastre 11/23/2009 09:16'!
testInsert

	| someUserIds |

	self resetStorage.
	self addDummyUsers: 10 times: 1.
	someUserIds := users collect:[:u| u id].

	pool readOnly:[
		ARDummyUser findAll do:[:guy|
			self assert: (someUserIds includes: guy id).
			self assert: (users anySatisfy: [:u| u id = guy id])]].
	
	pool readOnly:[
		self assert: ARDummyUser size = 10.
		self assert: (ARDummyUser sizeOfIndex: #id) = 10.
		self assert: (ARDummyUser sizeOfIndex: #username) = 10].
! !

!ARIndexingTest methodsFor: 'testing' stamp: 'sebastian.sastre 5/11/2010 23:30'!
testLocate

	self resetStorage.
	self addDummyUsers: 5 times: 1.

	pool readOnly:[
		ARDummyUser findAll do:[:guy|
			self assert: (ARDummyUser atId: guy id) = guy]].
	

	pool readOnly:[|some|
		some := ARDummyUser find: users anyOne createdOn of: #createdOn.
		"This may fail if the host is too slow (unable to store 5 in less than a sec)"
		self assert: some size = 5.].
! !

!ARIndexingTest methodsFor: 'testing' stamp: 'sebastian.sastre 11/24/2009 12:42'!
testNestedDynamicVariables

	self assert: ARCurrentClient value isNil.

	ARCurrentClient use: 1 during:[
		self assert: ARCurrentClient value = 1.
		ARCurrentClient use: 2 during:[
			self assert: ARCurrentClient value = 2].
		self assert: ARCurrentClient value = 1].
	
	self assert: ARCurrentClient value isNil

! !

!ARIndexingTest methodsFor: 'testing' stamp: 'seb 9/3/2011 23:54'!
testOccurrences

	| dates someUserIds sorted |

	self resetStorage.
	dates := Array with: TimeStamp today with: TimeStamp yesterday.
	self addDummyUsers: 10 times: 1 do:[:guy| guy createdOn: dates shuffled first].
	someUserIds := users collect:[:u| u id].
	sorted := users asSortedCollection:[:a :b| a createdOn > b createdOn].
	
	pool readOnly:[
		(ARDummyUser from: 1 to: 10 of: #createdOn) do:[:guy|
			self assert: (someUserIds includes: guy id).
			self assert: (users anySatisfy: [:u| u id = guy id])]].
	
	pool readOnly:[
		"All the occurrences of dates must fetch all the guys"
		self assert: (ARDummyUser from: 1 to: 10 of: #createdOn) size = users size.
		(ARDummyUser from: 1 to: 10 of: #createdOn) do:[:guy|
			self assert: (users anySatisfy: [:e| e = guy])]].

	pool readOnly:[|guys group todayPeople yesterdayPeople |
		"the occurrences must be fetched sorted"
		todayPeople := users select:[:e| e createdOn = TimeStamp today].
		yesterdayPeople := users select:[:e| e createdOn = TimeStamp yesterday].
		guys := ARDummyUser from: 1 to: yesterdayPeople size  of: #createdOn.
		self assert: guys size = yesterdayPeople size.
		self assert: (guys allSatisfy: [:e| e createdOn = TimeStamp yesterday]).

		guys := ARDummyUser from: 1 to: (yesterdayPeople size -1) of: #createdOn.
		self assert: guys size = (yesterdayPeople size -1).
		self assert: (guys allSatisfy: [:e| e createdOn = TimeStamp yesterday]).

		guys := ARDummyUser from: 2 to: yesterdayPeople size of: #createdOn.
		self assert: guys size = (yesterdayPeople size - 1).
		self assert: (guys allSatisfy: [:e| e createdOn = TimeStamp yesterday]).

		guys := ARDummyUser from: 2 to: (yesterdayPeople size -1) of: #createdOn.
		self assert: guys size = (yesterdayPeople size -2).
		self assert: (guys allSatisfy: [:e| e createdOn = TimeStamp yesterday]).

		guys := ARDummyUser from: yesterdayPeople size to: (yesterdayPeople size +1) of: #createdOn.
		self assert: guys size = 2.
		self assert: guys first createdOn = TimeStamp yesterday.
		self assert: guys second createdOn = TimeStamp today.

		guys := ARDummyUser from: (todayPeople size -1) to: 1 of: #createdOn.
		self assert: guys size = (todayPeople size -1).
		self assert: (guys allSatisfy: [:e| e createdOn = TimeStamp today]).

		guys := ARDummyUser from: todayPeople size to: 2 of: #createdOn.
		group := ARDummyUser from: 2 to: todayPeople size of: #createdOn.
		self assert: (guys allSatisfy: [:e| group includes: e]).
		self assert: (group allSatisfy: [:e| guys includes: e]).
		self assert: guys size = (todayPeople size - 1).
		self assert: (guys allSatisfy: [:e| e createdOn = TimeStamp today]).

		guys := ARDummyUser from: (todayPeople size -1) to: 2 of: #createdOn.
		self assert: guys size = (todayPeople size -2).
		self assert: (guys allSatisfy: [:e| e createdOn = TimeStamp today]).
		].
! !

!ARIndexingTest methodsFor: 'testing' stamp: 'sebastian.sastre 9/27/2010 16:37'!
testRemove

	| someUserIds someUsernames creations |

	self resetStorage.
	self addDummyUsers: 10 times: 1.
	someUserIds := users collect:[:u| u id].
	someUsernames := users collect:[:u| u username].
	creations := users collect:[:u| u createdOn].

	pool readOnly:[
		ARDummyUser findAll do:[:guy|
			self assert: (someUserIds includes: guy id).
			self assert: (users anySatisfy: [:u| u id = guy id])]].
	
	pool readOnly:[
		self assert: ARDummyUser size = 10.
		self assert: (ARDummyUser sizeOfIndex: #id) = 10.
		self assert: (ARDummyUser sizeOfIndex: #username) = 10].

	pool commit:[ARDummyUser findAll do:[:u| u destroy]].

	pool readOnly:[self assert: ARDummyUser findAll isEmpty].
	
	pool readOnly:[
		self deny: ARDummyUser hasAny.
		self assert: ARDummyUser size isZero.
		self assert: (ARDummyUser sizeOfIndex: #id) isZero.
		self assert: (ARDummyUser sizeOfIndex: #username) isZero.

		someUsernames do:[:username|
			self assert: (ARDummyUser atUsername: username) isNil].
		
		creations do:[:dateTime|
			self assert: (ARDummyUser withDateTime: dateTime) isEmpty].
		].
! !

!ARIndexingTest methodsFor: 'testing' stamp: 'sebastian.sastre 11/23/2009 09:15'!
testSize

	| someId someUsers |

	self resetStorage.

	pool readOnly:[
		self deny: ARDummyUser hasAny.
		self assert: ARDummyUser size isZero].

	self addDummyUsers: 10 times: 1.
	someUsers := users copy.
	pool readOnly:[
		self assert: ARDummyUser size = 10.
		self assert: (ARDummyUser sizeOfIndex: #id) = 10.
		self assert: (ARDummyUser sizeOfIndex: #username) = 10].

	pool commit:[
		someId := (someUsers remove: someUsers anyOne) id.
		(ARDummyUser atId: someId) destroy].
	pool readOnly:[
		self assert: ARDummyUser size = 9.
		self assert: (ARDummyUser sizeOfIndex: #id) = 9.
		self assert: (ARDummyUser sizeOfIndex: #username) = 9].
	
	pool commit:[
		1 to: 9 do:[:i|
			someId := (someUsers remove: someUsers anyOne) id.
			(ARDummyUser atId: someId) destroy.
			self assert: ARDummyUser size = (9 -i).
			self assert: (ARDummyUser sizeOfIndex: #id) = (9 -i).
			self assert: (ARDummyUser sizeOfIndex: #username) = (9 -i)]].

	pool readOnly:[
		self deny: ARDummyUser hasAny.
		self assert: ARDummyUser size isZero.
		self assert: (ARDummyUser sizeOfIndex: #id) isZero.
		self assert: (ARDummyUser sizeOfIndex: #username) isZero].
	
	
! !

!ARIndexingTest methodsFor: 'testing' stamp: 'sebastian.sastre 11/23/2009 17:36'!
testStoreNil

	| oldId newId |

	self resetStorage.
	self addDummyUsers: 10 times: 1.
	oldId := users anyOne id.

	self 
		should:[pool commit:[((pool root at: #storage) 
									indexFor: ARDummyUser  
									selector: #id)
										add: nil
										attributes: Dictionary new
										for: (pool root at: #storage)]] 
		raise: ARCantStoreNil.

	
! !

!ARIndexingTest methodsFor: 'testing' stamp: 'sebastian.sastre 11/26/2009 16:47'!
testToMany

	| changedTag removedTag |

	self resetStorage.
	self addDummyUsers: 1 times: 1.
	self addDummyTags: 10 times: 1.

	"Check insert"
	pool commit:[
		ARDummyUser first tags: tags; save].
	
	pool readOnly:[|guy|
		self assert: ARDummyUser first tags notEmpty.
		guy := ((pool root at: #storage) find: ARDummyUser where: #tags includes: tags anyOne) first.
		self assert: guy id = ARDummyUser first id.
		self assert: ARDummyUser first id = (ARDummyUser withTag: tags anyOne) first id].
	
	"Check update"
	pool commit:[|guy|
		guy := ARDummyUser first.
		removedTag := guy tags rehash remove: guy tags anyOne.		
		self deny: (guy tags includes: removedTag).
		guy save].
	
	pool readOnly:[|guy|
		self assert: ARDummyUser first tags notEmpty.
		self assert: ARDummyUser first tags size = 9.
		self assert: ((pool root at: #storage) find: ARDummyUser where: #tags includes: removedTag) isEmpty.
		self assert: (ARDummyUser withTag: removedTag) isEmpty.
		].
	
	"Check remove"
	
! !

!ARIndexingTest methodsFor: 'testing' stamp: 'seb 9/3/2011 20:31'!
testUnique

	| someUsers |

	self resetStorage.
	self addDummyUsers: 10 times: 1.
	someUsers := users copy asOrderedCollection.

	self should:[  
		pool commit:[|guy|
			guy := someUsers first fresh.
			guy username: someUsers last username.
			guy save]]
		raise: ARKeyAlreadyStored.

! !

!ARIndexingTest methodsFor: 'testing' stamp: 'sebastian.sastre 11/22/2009 15:40'!
testUpdates

	| someUsers oldUsernames newUsernames someNames |

	self resetStorage.
	self addDummyUsers: 10 times: 1.
	someUsers := users copy.

	oldUsernames := pool readOnly: [ARDummyUser findAll collect:[:e| e username]].
 	newUsernames := ((1 to: 10) collect:[:i| UUID new asString36]) asSet.
	someNames := newUsernames copy. "convenience to make a destructive interation"
	
	pool commit:[
		someUsers do:[:guy| |answeredGuy newName oldName|
			answeredGuy := ARDummyUser atId: guy id.
			self assert: answeredGuy notNil.
			self assert: answeredGuy id = guy id.
			self assert: answeredGuy username = guy username.
			self deny: (guy isIdenticalTo: answeredGuy).
			newName := someNames remove: someNames anyOne.
			oldName := answeredGuy username.
			answeredGuy username: newName; save.
			
			self assert: (ARDummyUser atUsername: oldName) isNil.
			self assert: (ARDummyUser atId: answeredGuy id) username = newName]].

	pool readOnly:[
		oldUsernames do:[:anOldName|
			self assert: (ARDummyUser atUsername: anOldName) isNil].
		newUsernames do:[:aNewName|
			self assert: (ARDummyUser atUsername: aNewName) notNil]].
! !

ARIndexingTest subclass: #ARTreeSetTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Test'!

!ARTreeSetTest methodsFor: 'accessing' stamp: 'sebastian.sastre 3/9/2011 20:08'!
someAlphanumerics
	
	^ (1 to: 1000) collect:[:i|
		UUID new asString36]! !

!ARTreeSetTest methodsFor: 'accessing' stamp: 'sebastian.sastre 3/9/2011 19:53'!
someNumbers

	| rnd |

	rnd := Random new.
	rnd seed: TimeStamp now asSeconds.
	
	^ (1 to: 1000) collect:[:i|
		rnd next]! !

!ARTreeSetTest methodsFor: 'testing' stamp: 'sebastiansastre 5/2/2012 14:16'!
testAlphanumeric
	
	| sorted tree elements |
	
	sorted := SortedCollection sortBlock: [:a :b| a < b].	
	elements := self someAlphanumerics.
	
	pool commit:[
		tree := ARTreeSet sortBy: #yourself.
		pool root at: #stuff put: tree.
		sorted addAll: elements.
		tree addAll: elements].

	1 to: elements size do:[:i|
		self assert: (sorted at: i) = (tree atIndex: i)].
	
	pool readOnly:[
		1 to: elements size do:[:i| |stuff|
			stuff := pool root at: #stuff.
			self assert: (sorted at: i) = (stuff atIndex: i)]]
		
! !

!ARTreeSetTest methodsFor: 'testing' stamp: 'sebastiansastre 5/2/2012 14:16'!
testNumbers
	
	| sorted tree elements |
	
	sorted := SortedCollection sortBlock: [:a :b| a < b].
	elements := self someNumbers.

	pool commit:[
		tree := ARTreeSet sortBy: #yourself.		
		pool root at: #stuff put: tree.
		sorted addAll: elements.
		tree addAll: elements].

	1 to: elements size do:[:i|
		self assert: (sorted at: i) = (tree atIndex: i)].
	
	pool readOnly:[
		1 to: elements size do:[:i| |stuff|
			stuff := pool root at: #stuff.
			self assert: (sorted at: i) = (stuff atIndex: i)]]
	
		
! !

!ARTreeSetTest methodsFor: 'testing' stamp: 'sebastiansastre 5/2/2012 14:17'!
testNumbersSmall
	
	| sorted tree elements |
	
	sorted := SortedCollection sortBlock: [:a :b| a < b].
	
	elements := #(
		0.000648575369570672
		0.000731434207750221
		0.001271925867196138
		0.001606590580943316
		0.003727566452570058
		0.00424666097585422
		).

	pool commit:[
		tree := ARTreeSet sortBy: #yourself.
		pool root at: #stuff put: tree.
		sorted addAll: elements.
		tree addAll: elements].

	1 to: elements size do:[:i|
		self assert: (sorted at: i) = (tree atIndex: i)].
	
	pool readOnly:[
		1 to: elements size do:[:i| |stuff|
			stuff := pool root at: #stuff.
			self assert: (sorted at: i) = (stuff atIndex: i)]]
	
		
! !

Object subclass: #ARAbstractIndexIterator
	instanceVariableNames: 'index'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Indices'!

!ARAbstractIndexIterator class methodsFor: 'initialization' stamp: 'sebastian.sastre 5/13/2010 14:31'!
on: anAttributeIndex

	^ self new 
		index: anAttributeIndex;
		yourself! !

!ARAbstractIndexIterator methodsFor: 'accessing' stamp: 'sebastian.sastre 5/13/2010 14:31'!
index
	^ index! !

!ARAbstractIndexIterator methodsFor: 'accessing' stamp: 'sebastian.sastre 5/13/2010 14:31'!
index: anObject
	index := anObject! !

ARAbstractIndexIterator subclass: #ARIndexIterator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Indices'!

!ARIndexIterator methodsFor: 'private' stamp: 'sebastian.sastre 5/13/2010 12:01'!
from: someSets addUpTo: endPosition on: someElements
	 "Adds to someElements the elements on the last set
	of someSets but limited to an endPosition (if reacheable)"
	
	| offset |
	
	"1. Calculate the size of the sum of all the sets 
	(drawing the segment helps to get why; it's about the offset)."
	offset := someSets inject: 0 into: [:sum :pair| sum + pair value].
	
	"2. Decide what to add regarding to endPosition 
	and the last set (which is not processed yet)"
	endPosition >= offset 
		ifTrue:[
			"All the elements of the last set must be added"
			someElements addAll: someSets last key]
		ifFalse:[|relativePosition firstSegmentOffset|
			"Some elements of last set must be added, so..
			calculate the relative position of the end of the iteration"
			firstSegmentOffset := someSets allButLast inject: 0 into: [:sum :pair| 
									sum + pair value].
			relativePosition := endPosition - firstSegmentOffset.
			"..and add that little bunch"
			someElements addAll: (someSets last key copyFrom: 1 to: relativePosition)]
	
! !

!ARIndexIterator methodsFor: 'queries' stamp: 'sebastian.sastre 5/13/2010 19:16'!
from: startPosition to: endPosition
	"Answers the aggregates found in the index 
	in the given interval (as ocurrences)"

	| tree sets current target offset relativeStart relativeEnd elements |
	
	tree := index indexingSupport.
	sets := OrderedCollection new.
	elements := TreeSet sortSelector: index selector equalitySelector: #isIdenticalTo:.

	"1. get the first element"
	current := tree getFirst value.
	"the sets are stored in association to nil
	but then nil will be its (cached) size"
	sets add: current -> current size.	

	"2. iterate until the target set is found.
	The target set is the one at which the element
	located at startPosition can be found."
	[target isNil] whileTrue:[
		(self inSegment: sets isIncluded: startPosition) 
			ifTrue:[target := current]
			ifFalse:[
				current := tree getNext value.
				sets add: current -> current size]].

	"3. Calculate the size of the sum of all the sets 
	previous to the target (drawing the segment 
	helps to get why; it's about the offset)."
	offset := sets allButLast inject: 0 into: [:sum :pair| sum + pair value].

	"4. Know the position to start fetching 
	relative to the target set."
	relativeStart := startPosition - offset.
	relativeStart isZero ifTrue: [
		relativeStart := 1].

	relativeEnd := sets last value min: (endPosition - offset).		

	"5. fetch the elements of the target set
	This gets discernable in 2 seconds if you sketch 
	the sgments."
	elements addAll: (target 
						copyFrom: relativeStart 
						to: relativeEnd).

	"6. if the endPosition is inside the target set then we
	can return the fetched elements by now."
	endPosition <= (sets inject: 0 into: [:sum :pair| sum + pair value]) ifTrue:[
		^ elements].

	"7. iterate until the set including the endPosition is found,
	adding elements as it goes."
	[sets last value <= endPosition] whileTrue:[
		current := tree getNext.
		current 
			ifNil:[
				"End of the index reached!!"
				^elements]
			ifNotNil:[current := current value].
		sets add: current -> current size.
		self from: sets addUpTo: endPosition on: elements].
	
	"8. Return the precious results :)"
	^ elements ! !

!ARIndexIterator methodsFor: 'testing' stamp: 'sebastian.sastre 5/13/2010 10:55'!
inSegment: someSets isIncluded: startPosition
	"Answers true if startPosition is reachable in the 
	segment represented by the sum of the sizes of someSets.
	Note: it caches in the value of the size of the sets in the 
	association so the expensive operation is made only once."
	
	^ startPosition <= (someSets inject: 0 into:[:sum :pair|
		pair value isNil ifTrue:[
			pair value: pair key size].
		sum + pair value])! !

ARAbstractIndexIterator subclass: #ARReverseIndexIterator
	instanceVariableNames: 'aggregateSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Indices'!

!ARReverseIndexIterator methodsFor: 'accessing' stamp: 'sebastian.sastre 5/13/2010 16:23'!
aggregateSize 

	^ aggregateSize ifNil:[self initializeAggregateSize]! !

!ARReverseIndexIterator methodsFor: 'queries' stamp: 'sebastian.sastre 5/13/2010 21:01'!
from: startPosition to: endPosition
	"Answers the aggregates found in the index 
	in the given interval (as ocurrences)"

	^ (ARIndexIterator on: index) 
		from: endPosition  
		to: startPosition ! !

!ARReverseIndexIterator methodsFor: 'initialization' stamp: 'sebastian.sastre 5/13/2010 18:52'!
initializeAggregateSize 

	^ aggregateSize := index aggregateClass size! !

Object subclass: #ARAggregate
	instanceVariableNames: 'id isNew oid isProxied'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Model'!

ARAggregate class
	instanceVariableNames: 'cache'!

!ARAggregate commentStamp: '<historical>' prior: 0!
An ARAggregate is a minimalistic abstraction of a unique persistent object stored in a storage (which currently is stored using an OmniBase repository).

The cache instance class variable will store transidently some aggregates that may be configured to do so (by how the query is implemented). That cache is designed to have an very short life span: the same duration of a transaction. When the transaction commits or aborts, those elements will be purged. Purging the cache is harmless for the transaction.

!

ARAggregate class
	instanceVariableNames: 'cache'!

!ARAggregate class methodsFor: 'private' stamp: 'sebastian.sastre 11/20/2009 10:14'!
addIndicesTo: aStorage
	"Add the indices of this aggregate to aStorage.
	Note: this can be used for tunning when developing.
	Not normally used in production (unless installing a 
	new repo is part of the app)."

	"we are using one container per index"
	self indices do:[:anIndex| 
		aStorage addIndex: anIndex]! !

!ARAggregate class methodsFor: 'actions' stamp: 'sebastian.sastre 4/25/2010 11:58'!
addSortSelector: aSymbol for: anIndexedSelector on: someIndices
	"Sets aSymbol in the index found at someIndices that is 
	working to keep fast access to anIndexedSelector 
	of this kind of aggregate."

	(someIndices detect: [:e| e indexesClass: self selector: anIndexedSelector]) sortSelector: aSymbol ! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 2/17/2010 12:50'!
aggregateAtId: anId
	"Answers the aggregate found at anId or nil.
	Note: 
	1. it allows you to search in an abstraction of the aggregate.
	2. is not proxied (trans-transactional)"

	^ (self atId: anId) aggregate ! !

!ARAggregate class methodsFor: 'accessing' stamp: 'sebastian.sastre 12/11/2010 11:52'!
aggregateClasses
	"Answers the set of classes which are concrete aggregates.
	Note: deprecated or abstract ones are going to be excluded"

	^ self withAllSubclasses reject: [ :aClass | aClass isAbstract or: [ aClass isDeprecated ] ]! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastiansastre 11/20/2011 14:50'!
all

	^ self findAll! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastiansastre 11/20/2011 14:11'!
atId: anId
	"Answers the aggregate found at anId or nil."

	|answer|

	self aggregateClasses do:[:anAggregateClass| |found|
		anAggregateClass == ARAggregate ifFalse:[
			found := anAggregateClass basicAtId: anId.
			found notNil ifTrue: [
				answer := found]]].
	^ answer! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastiansastre 9/30/2011 13:20'!
atIdOld: anId
	"Answers the aggregate found at anId or nil.
	Note: it allows you to search in an abstraction of the aggregate."

	|answer|

	self aggregateClasses do:[:anAggregateClass| |found|
		anAggregateClass == ARAggregate ifFalse:[
			found := anAggregateClass basicAtId: anId.
			found notNil ifTrue: [
				answer := found]]].
	^ answer! !

!ARAggregate class methodsFor: 'accessing' stamp: 'sebastian.sastre 12/16/2010 09:32'!
attributes
	"Answers the collection of instance variables 
	that (in this kind of aggregates) are publicly manipulable."
	
	^ self persistentlyMutableInstanceVariables ! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastian.sastre 12/23/2010 07:54'!
basicAtId: anId
	"Answers the aggregate found at anId"

	^ self storage find: self where: #id is: anId! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastian.sastre 4/25/2010 12:41'!
basicFindAll
	"Answers all stored instances of this kind aggregate
	including the ones of its subclasses."

	| subaggregateClasses |

	subaggregateClasses := self aggregateClasses.
	^  self basicFindAllFor: subaggregateClasses
! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 12/4/2011 21:04'!
basicFindAllFor: someAggregateClasses
	"Answers all stored instances of this kind aggregate
	including the ones of the kinds at someAggregateClasses."

	| answer index |
	
	index := someAggregateClasses
				detect:[:anAggregateClass| self storage hasIndicesFor: anAggregateClass]
				ifNone:[nil].
	index isNil
		ifTrue:[answer := TreeSet sortBy: #id]
		ifFalse:[
			index := self storage indexFor: index selector: #id.
			index hasSortSelector 
				ifTrue:[answer := TreeSet sortBy: index sortSelector]
				ifFalse:[answer := TreeSet sortBy: #id]].

	someAggregateClasses do:[:anAggregateClass|
		anAggregateClass == ARAggregate ifFalse:[
			answer addAll: anAggregateClass localFindAll]].	 
	
	^ answer ! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 11/20/2011 14:25'!
basicSizeOfIndex: aSymbol
	"Answers the quantity of stored instances 
	of this particular aggregate on the aSymbol index"

	^ (self storage indexFor: self selector: aSymbol) size
! !

!ARAggregate class methodsFor: 'accessing' stamp: 'sebastian.sastre 11/25/2009 16:29'!
cache
	"Answers the cache that relates query results to transactions.
	Remember: this are very transident, very short term kind of cache.
	It's only valid until the transaction finished (one way or another)"
	^ cache ifNil:[self initializeCache]! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 14:37'!
cardKeySize

	^  32! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastian.sastre 11/19/2009 21:31'!
currentTransaction

	^ OmniBase currentTransaction! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 14:19'!
dateKeySize

	^ 8! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 14:20'!
defaultKeySize
	"Answers the default size for keys on indices.
	It assumes it's about an UUID value asString36 which
	seems to need 25 bytes, so we're being wasty in 1 byte just in case..

	self assert: (UUID new asString36 asBtreeKeyOfSize: 26) last = 0"
	
	^ self idKeySize! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 14:31'!
emailKeySize

	^ 40! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 1/10/2012 11:53'!
ensureValueOf: attribute in: anAggregate from: anEditedAggregate
	"Ensures the changes of anEditedAggregate to be present in anAggregate.
	Note that we assume that every instVar of an aggregate has its ideal accessor (and mutator).
	We count on that for getting the fresh value and setting it.
	Beside, we'll check if those instVars are holding a collection to retrieve a fresh collected version of its elements."

	| newValue |
	
	newValue := anEditedAggregate perform: attribute asSymbol. 
	(newValue isCollection and:[
	newValue deservesPersistentRead])
		ifFalse:[ anAggregate perform: (attribute,':') asSymbol with: newValue fresh ]
		ifTrue:[|freshValues|
			"This assumes that you are using simple collections in aggregate instVars (that can normally perform a #collect:)"
			anAggregate 
				perform: (attribute,':') asSymbol 
				with: (newValue collect: [:e| |freshOne|
						freshOne := e fresh.
						freshOne ifNil:[self error: 'You probably should have saved ', e asString,' before trying to save ', anAggregate asString].
						freshOne])].
! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 1/10/2012 11:54'!
ensureValuesIn: anAggregate from: anEditedAggregate
	"Ensures the changes of anEditedAggregate to be present in anAggregate.
	Note that we assume that every instVar of an aggregate has its ideal accessor (and mutator).
	We count on that for getting the fresh value and setting it.
	Beside, we'll check if those instVars are holding a collection to retrieve a fresh collected version of its elements."

	anAggregate class persistentlyMutableInstanceVariables do:[:attribute|
			self ensureValueOf: attribute in: anAggregate from: anEditedAggregate]
	
! !

!ARAggregate class methodsFor: 'testing' stamp: 'sebastian.sastre 11/22/2009 14:56'!
existsId: anId
	"Answers true if there is an object stored at anId.
	Note: it allows you to search in an abstraction of the aggregate."

	^ (self atId: anId) notNil! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 23:21'!
find: aValue of: aSelector
	"Answers the aggregate (or aggregates) found for aValue indexed by aSelector"
	
	^ self storage find: self  where: aSelector is: aValue! !

!ARAggregate class methodsFor: 'queries' stamp: 'seb 9/4/2011 00:17'!
findAll
	"Answers all instances of this kind (and sub kinds) of aggregate or an empty set.
	Note 1: this is only a good idea when you have stored few (less than 100?)
	and starts to become a REALLY BAD idea when you have, for instance, more than 100 thounsand.
	Note 2: if this aggregate has subclasses their instances are going to be included."

	| cached answer |

	"First try to do a cache hit, if is not a hit then 
	lazily store the results of a gentle fallback to the real query"
	"cached := self cache at: self currentTransaction named: 'findAll'."	

	^ self basicFindAll

	"Answers a copy (using asOrderedCollection) 
	so the application can do its stuff without 
	comprimising the integrity of the cache"
"	^ cached isNil
		ifFalse:[cached asOrderedCollection]
		ifTrue:[
			answer := self basicFindAll.
			self cache 
					at: self currentTransaction
					put: answer
					named: 'findAll'.
			answer asOrderedCollection]."

! !

!ARAggregate class methodsFor: 'private' stamp: 'seb 9/4/2011 00:03'!
findAllOn: anIndexedSelector

	^ (self storage isIndexing: self with: anIndexedSelector)
		ifTrue: [(self storage indexFor: self selector: anIndexedSelector) values]
		ifFalse: [self error: self name, ' is not being indexed with ', anIndexedSelector]
	 
	! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 12/17/2009 12:51'!
findAllSortedBy: aSelector
	"Answers all instances or ans empty set.
	Note: this is only a good idea when you have very few of this objects
	and starts to become a REALLY BAD idea when you have, for instance, more than 100."

	|answer|

	answer := TreeSet sortSelector: aSelector equalitySelector: #isIdenticalTo:.
	self aggregateClasses do:[:anAggregateClass|
		anAggregateClass == ARAggregate ifFalse:[
			answer addAll: anAggregateClass localFindAll]].

	^ answer! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 2/16/2010 11:04'!
first
	"Answers the first of the stored instances 
	of this particular aggregate on the primary index"

	^ self primaryIndex first 
! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 19:21'!
first: anInteger

	^ self first: anInteger of: #id! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 19:21'!
first: anInteger of: aSelector
	"Answers the first anInteger aggregates indexed by aSelector"
	
	^ (self storage indexFor: self selector: aSelector) first: anInteger! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 2/16/2010 11:05'!
firstAggregate
	"Answers the first of the stored instances 
	of this particular aggregate ensuring is not proxied"

	^ self first aggregate 
! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 15:46'!
floatKeySize

	^ 6! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 22:18'!
from: start to: end of: aSelector
	"Answers the aggregates indexed by aSelector
	that are from the occurrence number start 
	up to the occurrence number end"
	
	^ self storage find: self from: start to: end of: aSelector
	
	
! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 21:55'!
fromPosition: start to: end of: aSelector
	"Answers the aggregates indexed by aSelector
	that are from the occurrence number start up to the occurrence number end"
	
	^ (self storage indexFor: self selector: aSelector) fromPosition: start to: end! !

!ARAggregate class methodsFor: 'testing' stamp: 'sebastian.sastre 11/23/2009 09:14'!
hasAny
	"Answers true if this aggregate has at least one stored aggregate."
	^ self size > 0! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 14:20'!
idKeySize
	"Answers the default size for keys on indices.
	It assumes it's about an UUID value asString36 which
	seems to need 25 bytes, so we're being wasty in 1 byte just in case..

	self assert: (UUID new asString36 asBtreeKeyOfSize: 26) last = 0"

	^ 26! !

!ARAggregate class methodsFor: 'accessing' stamp: 'sebastiansastre 10/17/2011 16:29'!
indexOn: aSelectorSymbol

	^ self storage indexFor: self selector: aSelectorSymbol
! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 14:37'!
indices
	"Answers the indices for this active record 
	so it performs well for the application using it.
	Note: repository rebuild is recomended"

	"The id is a required index to locate instances by id"
	^ OrderedCollection new
		add: (ARUniqueAttributeIndex new
				dontAllowNil;
				keySize: self idKeySize;
				aggregateClass: self; 
				selector: #id;
				yourself);
		yourself! !

!ARAggregate class methodsFor: 'initialization' stamp: 'sebastian.sastre 11/25/2009 17:06'!
initialize

	Smalltalk addToStartUpList: self.
	Smalltalk addToShutDownList: self! !

!ARAggregate class methodsFor: 'initialization' stamp: 'sebastian.sastre 11/25/2009 16:30'!
initializeCache
	"Answers the cache that relates query results to transactions.
	Remember: this are very transident, very short term kind of cache.
	It's only valid until the transaction finished (one way or another)"
	^ cache := ARCache new! !

!ARAggregate class methodsFor: 'testing' stamp: 'sebastian.sastre 11/20/2009 14:30'!
isAbstract
	"Answers true if the aggregate is abstract.
	Very useful to subclassify a family of aggregates but
	that it wouldn't make sense to create instances.

	You can (sub) implement where nessesary in an abstract aggregate like this:
	
	^ self == AnAbstractAggregateClass"
	
	^ false! !

!ARAggregate class methodsFor: 'testing' stamp: 'sebastian.sastre 11/20/2009 14:30'!
isDeprecated
	"Answers true if the aggregate is deprecated.
	Very useful for migrations after design changes.

	You can (sub) implement something like this in a deprecated class:
	
	^ self == ADeprecatedAggregateClass"
		
	^ false! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 15:40'!
largePositiveIntegerKeySize

	^ 8! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 19:21'!
last: anInteger

	^ self last: anInteger of: #id! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 19:21'!
last: anInteger of: aSelector
	"Answers the last anInteger aggregates indexed by aSelector"
	
	^ (self storage indexFor: self selector: aSelector) last: anInteger! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastian.sastre 5/11/2010 17:00'!
localFindAll
	"Answers all stored instances or an empty set
	for this aggregate alone without including 
	the ones of its subclasses."

	^ (self storage isIndexing: self with: #id)
		ifTrue: [self primaryIndex indexingSupport values]
		ifFalse: [#()]
	 
	! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 15:35'!
maxKeySize

	^  255! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 15:55'!
minKeySize

	^  5! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 14:32'!
nameKeySize

	^ 64! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastian.sastre 12/23/2010 11:13'!
odbTransientInstanceVariables
	"This method tells OmniBase which instance variables should not be stored into the database."

	"the hashCache is transident stuff for the persistence machinery"
	^super odbTransientInstanceVariables
		add: 'isProxied';  "<--- isProxied becomes true once the persistent got rid of the proxies (and subproxies)"
		yourself! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastian.sastre 12/13/2010 13:17'!
persistentInstanceVariables 
	"Answers the instance variables in the receiver that are going to be persisted (not ignored by the storage.)"
	
	^ self allInstVarNames reject:[:instVarName|
		self odbTransientInstanceVariables includes: instVarName]! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastian.sastre 12/16/2010 09:30'!
persistentlyMutableInstanceVariables
	"Answers the instance variables of this aggregate that are going to 
	be publicly manipulable."
	
	"We exclude #isNew and the #id of the aggregate"
	^  self persistentInstanceVariables reject:[:e|
		e = 'id' or:[
		e = 'isNew']]! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastian.sastre 11/23/2009 11:45'!
primaryIndex
	"Answers primary index (which index identity)
	of this particular kind of aggregates"

	^ self storage indexFor: self selector: #id
! !

!ARAggregate class methodsFor: 'actions' stamp: 'sebastian.sastre 11/25/2009 16:24'!
reset
	"Resets the cache of this active recod (alone)"
	
	cache isNil ifFalse: [|exCache|
		exCache := cache.
		cache := nil.
		exCache purge]
	! !

!ARAggregate class methodsFor: 'actions' stamp: 'sebastian.sastre 12/15/2010 13:00'!
resetAll
	"Reset the caches of all the subhierarchy"
	self withAllSubclasses do:[:anAggregateClass|
		anAggregateClass reset]! !

!ARAggregate class methodsFor: 'actions' stamp: 'sebastian.sastre 11/25/2009 17:06'!
shutDown

	self resetAll ! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 11/22/2009 15:12'!
size
	"Answers the quantity of aggregates of this kind"

	^ self sizeOfIndex: #id

! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 11/30/2009 09:17'!
sizeOfIndex: aSymbol
	"Answers the quantity of aggregates of this kind on the requested index"

	|answer|
	answer := 0.
	self aggregateClasses do:[:anActiveRecordClass|
		anActiveRecordClass == ARAggregate ifFalse:[
			answer := answer + (anActiveRecordClass basicSizeOfIndex: aSymbol)]].

	^ answer


! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 14:28'!
smallIntegerKeySize

	^ 5! !

!ARAggregate class methodsFor: 'actions' stamp: 'sebastian.sastre 11/25/2009 17:07'!
startUp

	self resetAll ! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastian.sastre 11/20/2009 14:31'!
storage

	^ self currentTransaction root at: #storage! !

!ARAggregate class methodsFor: 'queries' stamp: 'sebastian.sastre 2/7/2011 01:38'!
uncachedFindAll

	^ self basicFindAll! !

!ARAggregate class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 14:32'!
usernameKeySize

	^ 20! !

!ARAggregate methodsFor: 'testing' stamp: 'sebastian.sastre 2/16/2010 11:44'!
= anAggregate

	^ anAggregate isAggregate and:[
	id = anAggregate id]! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastiansastre 11/19/2011 09:57'!
aggregate
	"Answers the receiver.
	Applicability: useful to make proxies fail into the real persistent"

	"This is to save us from redundant deep introspections (to de-proxy this aggregate)"
	^ (isProxied isNil or:[
	isProxied])
		ifFalse:[self]  
		ifTrue:[|real|
			real := self persistentRead.
			isProxied := false.
			real]! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastiansastre 9/25/2011 15:42'!
asBtreeKeyOfSize: keySize
	keySize < self class idKeySize ifTrue: [self error: 'Key size is too small'].

	^ id asBtreeKeyOfSize: keySize ! !

!ARAggregate methodsFor: 'private' stamp: 'sebastiansastre 3/6/2012 16:49'!
asDictionary 

	| dic |
	
	dic := Dictionary new.
	
	self class allInstVarNames do:[:name|
		dic at: name put: (self instVarNamed: name) asString].
	
	^ dic! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastiansastre 9/30/2011 17:40'!
basicFresh
	"Answers a fresh version of this aggregate.
	Fresh meaning that it's just read from disk (read and reified)."

	^ self isNew 
		ifTrue: [self]
		ifFalse: [self currentTransaction freshObjectAt: self odbGetOID] ! !

!ARAggregate methodsFor: 'private' stamp: 'sebastian.sastre 1/8/2011 12:16'!
basicPrintOn: aStream
	
	aStream nextPutAll: '(',id asString,')'! !

!ARAggregate methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 19:44'!
beDeleted
	"Mark the receiver as deleted"
	self error: 'Aggregates can''t be deleted (only destroyed)'! !

!ARAggregate methodsFor: 'private' stamp: 'sebastiansastre 3/2/2012 11:33'!
beNew
	"Mark the receiver as new.
	WARNING: use it very wisely. 
	When saved this forces active records to be added in the storage.
	I've added this feature to be useful to migrate from 
	whatever source to a brand new repository"
	isNew := true.
	oid := nil.! !

!ARAggregate methodsFor: 'private' stamp: 'sebastian.sastre 3/10/2011 16:07'!
beNotNew
	"Mark the receiver as NOT new.
	WARNING: meant to be private.
	Use it very wisely."

	isNew := false! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 20:42'!
container
	"Answers the container where the receiver should be stored"
	
	^ self storage containerFor: self! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastiansastre 10/20/2011 08:53'!
current
	"Answers a version of this aggregate in the current transaction."

	^ self fresh! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastiansastre 10/5/2011 11:59'!
currentAt
	"Answers this aggregate in the current transaction.
	Note hat, unless this isNew,  it's fetched by the #id index"

	^ self isNew 
		ifTrue: [self]
		ifFalse: [self class atId: self id]! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 21:30'!
currentTransaction

	^ self class currentTransaction! !

!ARAggregate methodsFor: 'actions' stamp: 'sebastiansastre 6/15/2011 21:24'!
deleteFiles
	"Deletes files in the directory where this aggregate is supposed to store attached files"
	
	| directory |
	
	directory := self session pathFor: self.

	directory exists ifTrue:[
		directory deleteAll]
! !

!ARAggregate methodsFor: 'actions' stamp: 'sebastiansastre 9/29/2011 20:19'!
destroy
	"The receiver is going to be permanently removed from the repository.
	CAUTION: there is no undo for this action and it may leave inconsistences in the repository.
	Please be extremely aware of what you are doing when asking an aggregate to be destroyed."

	| persistent |	
	self onAboutToDestroy.

	self isNew ifTrue:[^self onAfterDestroy].

	self isODBPersistent
		ifTrue:[persistent := self]
		ifFalse:[persistent := self fresh].

	persistent ifNotNil:[
		self storage remove: persistent.
		persistent onAfterDestroy]! !

!ARAggregate methodsFor: 'actions' stamp: 'sebatian.sastre 5/1/2010 15:12'!
destroyIfStored
	"Destroys this aggregate only 
	if it was previously stored."
	
	self onAboutToDestroy.

	self isNew ifTrue:[^self].

	[:storedOrNil|
		storedOrNil ifNotNil:[
			self storage remove: storedOrNil]] value: self fresh	realPersistent
	! !

!ARAggregate methodsFor: 'testing' stamp: 'sebastian.sastre 3/25/2011 12:72'!
exists
	"Answers true if this aggregate is "
	^ self class existsId: self id! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastiansastre 10/20/2011 08:53'!
fresh
	"Answers this aggregate taken from the current transaction."

	^ self isNew 
		ifTrue: [self]
		ifFalse: [|theOid|
			theOid := self odbGetOID.
			theOid isNil 
				ifFalse: [self currentTransaction objectAt: theOid]
				ifTrue:[self freshAt]] ! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastiansastre 10/20/2011 08:53'!
freshAt
	"Answers this aggregate in the current transaction.
	Note hat, unless this isNew,  it's fetched by the #id index"

	^ self isNew 
		ifTrue: [self]
		ifFalse: [self class atId: self id]! !

!ARAggregate methodsFor: 'testing' stamp: 'sebastiansastre 9/30/2011 11:37'!
hasOid

	^ oid notNil! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastian.sastre 11/27/2009 12:02'!
hash
	
	^ id hash! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 19:32'!
id
	^ id! !

!ARAggregate methodsFor: 'initialization' stamp: 'sebastiansastre 3/2/2012 11:33'!
initialize

	super initialize.

	id := UUID new asString36.
	isNew := true.
	oid := nil.! !

!ARAggregate methodsFor: 'initialization' stamp: 'sebastian.sastre 11/26/2009 17:32'!
initializeIsNew
	"Answers true if the receiver is an aggregate not saved yet)"
	^ isNew := self odbGetOID notNil! !

!ARAggregate methodsFor: 'testing' stamp: 'sebastian.sastre 2/16/2010 11:44'!
isAggregate 

	^ true! !

!ARAggregate methodsFor: 'testing' stamp: 'sebastian.sastre 11/26/2009 17:48'!
isNew
	"Answers true if the receiver is an aggregate not saved yet."
	
	^ isNew! !

!ARAggregate methodsFor: 'private' stamp: 'sebastiansastre 3/6/2012 16:48'!
jsonOn: aStream

	self asDictionary jsonOn: aStream! !

!ARAggregate methodsFor: 'actions' stamp: 'sebastiansastre 10/4/2011 20:27'!
makeObjectID
	"Answers the fabricated instance of ODBObjectID 
	given the container and index data of this persistent aggregate."
	
	^ oid notNil 
		ifTrue:[ODBObjectID containerID: oid key index: oid value]
		ifFalse:[nil]! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastiansastre 10/5/2011 11:12'!
odbGetOID
	"Answers the omnibase object ID for this persistent object.
	Note that this overrides OmniBase default way to get the object id by:
	1. trying  the 'fabricated' object ID using the container and index data, BWT ~x8 faster than
	2. falling back to the OmniBase's way"
		
	^ self hasOid 
		ifTrue:[self makeObjectID]
		ifFalse:[super odbGetOID]
	
! !

!ARAggregate methodsFor: 'private' stamp: 'sebastian.sastre 3/10/2011 16:06'!
odbMadePersistentIn: anOmniBaseTransaction
	"This method is sent when the object is made persistent."

	self markDirty. "may be redundant, but you know.."
	
	super odbMadePersistentIn: anOmniBaseTransaction! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastiansastre 9/30/2011 11:07'!
oid
	^ oid! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastiansastre 3/3/2012 10:12'!
oid: anObject

	"Note: reject setting nil."
	(oid notNil and:[
	anObject isNil]) ifTrue:[^self].

	"(oid notNil and:[
	anObject notNil]) ifTrue:[self halt]."

	"oid ifNotNil:[
		self assert: anObject notNil]."
	
	oid := anObject! !

!ARAggregate methodsFor: 'reactions' stamp: 'sebastian.sastre 11/19/2009 19:53'!
onAboutToDestroy
	"This aggregate is about to be permanently removed from the repository.
	CAUTION: there is NO UNDO for this action"
	
! !

!ARAggregate methodsFor: 'reactions' stamp: 'sebastian.sastre 3/10/2011 10:06'!
onAboutToEnsureValuesIn: aFreshAggregate from: anEditedAggregate
	"aFreshAggregate about to have the values of anEditedAggregate
	just before updating it into the storage."
	
	"no-op"! !

!ARAggregate methodsFor: 'reactions' stamp: 'sebastian.sastre 3/10/2011 10:05'!
onAboutToInsert
	"The receiver is about to be saved for the first time."
	
	"no-op"! !

!ARAggregate methodsFor: 'reactions' stamp: 'sebastiansastre 9/30/2011 10:49'!
onAboutToRefresh
	"The receiver is about to be refreshed from the storage."
	
	"no-op"
	! !

!ARAggregate methodsFor: 'reactions' stamp: 'sebastiansastre 10/17/2011 21:44'!
onAboutToSave
	"The receiver is about to be saved."
	
	"no-op"
	
! !

!ARAggregate methodsFor: 'reactions' stamp: 'sebastian.sastre 3/10/2011 10:11'!
onAfterDestroy
	"This aggregate was just removed (permanently) from the storage."
	
	"no-op"
	
! !

!ARAggregate methodsFor: 'reactions' stamp: 'sebastiansastre 10/20/2012 17:11'!
onAfterInsertion
	"The receiver was just saved for the first time."
	
	"no-op"! !

!ARAggregate methodsFor: 'reactions' stamp: 'sebastian.sastre 3/10/2011 10:54'!
onAfterSave
	"This aggregate was just saved (and it wasn't saved for the first time)."
	
	"no-op"! !

!ARAggregate methodsFor: 'private' stamp: 'sebastian.sastre 1/8/2011 12:17'!
printOn: aStream

	super printOn: aStream.

	self basicPrintOn: aStream ! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastiansastre 9/30/2011 17:40'!
read
	"Answers a fresh version of this aggregate.
	Fresh meaning that it's just read from disk."

	^ self isNew 
		ifTrue: [self]
		ifFalse: [self currentTransaction freshObjectAt: self odbGetOID] ! !

!ARAggregate methodsFor: 'actions' stamp: 'sebastiansastre 9/30/2011 10:49'!
refresh 
	"Makes the receiver to refresh itself.
	At first glance this seems silly but it absolutely 
	makes sense when this aggregate is composed 
	by another aggregate/s."
	
	self onAboutToRefresh.
	
	self onAboutToEnsureValuesIn: self from: self.
	self class ensureValuesIn: self from: self
! !

!ARAggregate methodsFor: 'actions' stamp: 'sebastiansastre 10/20/2012 17:13'!
save
	"Saves this Aggregate in the repository"

	| persistent |

	self onAboutToSave.

	isNew
		ifTrue:[
			self onAboutToInsert.
			self refresh.
			isNew := false.
			self storage add: self.
			self onAfterInsertion]
		ifFalse:[
			self isODBPersistent
				ifTrue:[persistent := self]
				ifFalse:[persistent := self current].
			self onAboutToEnsureValuesIn: persistent from: self.
			self class ensureValuesIn: persistent from: self.
			isNew := false.
			self currentTransaction markDirty: persistent.
			self storage update: persistent.
			persistent onAfterSave]! !

!ARAggregate methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 19:41'!
saveIfStored
	"Saves the receiver only if the aggregate 
	was previously stored."
	
	self isNew ifFalse:[
		self save]! !

!ARAggregate methodsFor: 'actions' stamp: 'seb 9/3/2011 20:25'!
saveOld
	"Saves this ActiveRecord in the repository"
	
	self onAboutToSave.

	isNew
		ifTrue:[
			self refresh.
			self onAboutToInsert.
			self storage add: self.
			self onAfterInsertion.]
		ifFalse:[|persistent|
			persistent := self class atId: self id.
			self onAboutToEnsureValuesIn: persistent from: self.
			self class ensureValuesIn: persistent from: self.
			self currentTransaction markDirty: persistent.
			self storage update: persistent realPersistent.
			self onAfterSave]! !

!ARAggregate methodsFor: 'private' stamp: 'sebastiansastre 10/17/2011 22:11'!
setOid
	"Sets the object id value of this persistent object in the omnibase repository container.
	This message assumes that it has been made persistent in the current transaction."
	
	| theOid |

	theOid := super odbGetOID.
	self assert: theOid notNil.

	oid := theOid containerID -> theOid index! !

!ARAggregate methodsFor: 'private' stamp: 'sebastiansastre 3/3/2012 10:13'!
sixxIgnorableInstVarNames
	"If we are using SIXX to dump/import data (and you should!!), 
	then we will ignore the oid that was given to this aggregate in this (current) repository,
	otherwise, any import (which creates a new repo) would get confused."
	 
	^#(#oid)! !

!ARAggregate methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 19:59'!
storage

	^ self class storage! !

ARAggregate subclass: #ARPersistentModel
	instanceVariableNames: 'createdOn updatedOn isDeleted'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Model'!

ARPersistentModel subclass: #ARDummyPerson
	instanceVariableNames: 'firstName lastName tags'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Model'!

!ARDummyPerson class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 14:37'!
indices
	"Answers the indices for this active record 
	so it performs well for the application using it.
	Note: repository rebuild is recomended"

	^ super indices
		add: (ARAttributeIndex new 
				aggregateClass: self; 
				selector: #firstName;
				keySize: self nameKeySize;
				yourself);
		yourself! !

!ARDummyPerson class methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 17:38'!
withFirstName: aName

	^ self storage find: self where: #firstName is: aName! !

!ARDummyPerson methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 21:34'!
firstName
	^ firstName! !

!ARDummyPerson methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 21:34'!
firstName: anObject
	firstName := anObject! !

!ARDummyPerson methodsFor: 'initialization' stamp: 'sebastian.sastre 4/26/2010 18:26'!
initializeTags

	^ tags := Set new! !

!ARDummyPerson methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 21:34'!
lastName
	^ lastName! !

!ARDummyPerson methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 21:34'!
lastName: anObject
	lastName := anObject! !

!ARDummyPerson methodsFor: 'accessing' stamp: 'sebastian.sastre 4/26/2010 18:25'!
tags
	^ tags ifNil:[self initializeTags]! !

!ARDummyPerson methodsFor: 'accessing' stamp: 'sebastian.sastre 4/26/2010 18:25'!
tags: anObject
	tags := anObject! !

ARPersistentModel subclass: #ARDummyTag
	instanceVariableNames: 'label'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Model'!

!ARDummyTag class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 14:38'!
indices
	"Answers the indices for this active record 
	so it performs well for the application using it.
	Note: repository rebuild is recomended"

	^ super indices
		add: (ARAttributeIndex new 
				aggregateClass: self; 
				selector: #createdOn;
				keySize: self dateKeySize;
				yourself);
		yourself! !

!ARDummyTag class methodsFor: 'accessing' stamp: 'sebastian.sastre 4/26/2010 21:05'!
withLabel: aString

	^ self findAll 
		detect:[:e| e label = aString]
		ifNone:[nil]! !

!ARDummyTag methodsFor: 'initialization' stamp: 'sebastian.sastre 11/26/2009 11:40'!
initialize

	super initialize.

	label := UUID new asString36 ! !

!ARDummyTag methodsFor: 'accessing' stamp: 'sebastian.sastre 11/26/2009 11:34'!
label
	^ label! !

!ARDummyTag methodsFor: 'accessing' stamp: 'sebastian.sastre 11/26/2009 11:34'!
label: anObject
	label := anObject! !

ARPersistentModel subclass: #ARDummyUser
	instanceVariableNames: 'username email person tags'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Model'!

!ARDummyUser class methodsFor: 'queries' stamp: 'sebastian.sastre 11/23/2009 17:30'!
atUsername: anUsername

	^ self storage find: self where: #username is: anUsername! !

!ARDummyUser class methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 14:39'!
indices
	"Answers the indices for this active record 
	so it performs well for the application using it.
	Note: repository rebuild is recomended"

	^ super indices
		add: (ARAttributeIndex new 
				aggregateClass: self; 
				selector: #createdOn;
				keySize: self dateKeySize;
				yourself);
		add: (ARUniqueAttributeIndex new 
				dontAllowNil;
				aggregateClass: self; 
				selector: #username;
				keySize: self usernameKeySize;
				yourself);
		add: (ARManyAttributesIndex new 
				allowNil;
				aggregateClass: self; 
				selector: #tags;
				keySize: self idKeySize;				
				yourself);
		yourself! !

!ARDummyUser class methodsFor: 'queries' stamp: 'sebastian.sastre 9/27/2010 16:36'!
withDateTime: aDateTime

	^ self storage find: self where: #createdOn is: aDateTime! !

!ARDummyUser class methodsFor: 'queries' stamp: 'sebastian.sastre 11/26/2009 15:29'!
withTag: aTag

	^ self storage find: self where: #tags includes: aTag! !

!ARDummyUser methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 21:35'!
email
	^ email! !

!ARDummyUser methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 21:35'!
email: anObject
	email := anObject! !

!ARDummyUser methodsFor: 'testing' stamp: 'sebastian.sastre 4/26/2010 18:04'!
hasPerson

	^ person notNil ! !

!ARDummyUser methodsFor: 'testing' stamp: 'sebastian.sastre 11/26/2009 12:18'!
hasTags

	^ tags notNil and:[
	tags notEmpty]! !

!ARDummyUser methodsFor: 'initialization' stamp: 'sebastian.sastre 11/20/2009 11:09'!
initialize

	super initialize.

	username := UUID new asString36 ! !

!ARDummyUser methodsFor: 'initialization' stamp: 'sebastian.sastre 11/26/2009 16:47'!
initializeTags

	^ tags := Set new! !

!ARDummyUser methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 21:35'!
person
	^ person! !

!ARDummyUser methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 21:35'!
person: anObject
	person := anObject! !

!ARDummyUser methodsFor: 'accessing' stamp: 'sebastian.sastre 11/26/2009 13:26'!
tags
	^ tags
	"^ tags ifNil:[self initializeTags]"	! !

!ARDummyUser methodsFor: 'accessing' stamp: 'sebastian.sastre 11/26/2009 11:44'!
tags: anObject
	tags := anObject! !

!ARDummyUser methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 21:34'!
username
	^ username! !

!ARDummyUser methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 21:34'!
username: anObject
	username := anObject! !

!ARPersistentModel class methodsFor: 'queries' stamp: 'sebastiansastre 1/10/2012 11:56'!
all
	"Even deleted ones"
	^ super findAllOn: #id! !

!ARPersistentModel class methodsFor: 'queries' stamp: 'sebastiansastre 12/4/2011 21:08'!
findAll
	"All subclasses should expect that a #findAll means the not deleted guys."

	^ super findAll reject: [:each | each isDeleted ]! !

!ARPersistentModel class methodsFor: 'testing' stamp: 'sebastian.sastre 12/4/2009 10:17'!
isAbstract
	"Answers true if the aggregate is abstract.
	Very useful to subclassify a family of aggregates but
	that it wouldn't make sense to create instances.

	You can (sub) implement where nessesary in an abstract aggregate like this:
	
	^ self == AnAbstractAggregateClass"
	
	^ self == ARPersistentModel ! !

!ARPersistentModel methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 19:41'!
beDeleted
	"Mark the receiver as deleted"
	isDeleted := true! !

!ARPersistentModel methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 19:30'!
createdOn
	^ createdOn! !

!ARPersistentModel methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 19:40'!
createdOn: anObject
	createdOn := anObject! !

!ARPersistentModel methodsFor: 'actions' stamp: 'sebastiansastre 10/13/2011 09:19'!
delete
	"The receiver is going to be marked as deleted in the repository.
	Usually a safe operation: the aggregate will still be undeletable.
	This is for usability reasons. Use to destroy to vanish the aggregate."

	| persistent |
	
	self onAboutToDelete.

	self isNew
		ifTrue:[
			self refresh.
			isDeleted := true.
			self storage add: self]
		ifFalse:[
			self isODBPersistent
				ifTrue:[persistent := self]
				ifFalse:[persistent := self fresh.
						self onAboutToEnsureValuesIn: persistent from: self.
						self class ensureValuesIn: persistent from: self].
			persistent beDeleted.
			self currentTransaction markDirty: persistent.
			self storage update: persistent.
			persistent onAfterDelete]			
! !

!ARPersistentModel methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 19:55'!
deleteIfStored
	"Deletes this aggregate only 
	if it was previously stored."
	
	self isNew ifFalse:[
		self delete]! !

!ARPersistentModel methodsFor: 'initialization' stamp: 'sebastian.sastre 11/19/2009 19:33'!
initialize

	super initialize.

	createdOn := updatedOn := TimeStamp now.
	isDeleted := false.
! !

!ARPersistentModel methodsFor: 'testing' stamp: 'sebastiansastre 12/4/2011 21:08'!
isDeleted

	^ isDeleted notNil and:[
	isDeleted]! !

!ARPersistentModel methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 19:40'!
isDeleted: anObject
	isDeleted := anObject! !

!ARPersistentModel methodsFor: 'private' stamp: 'sebastian.sastre 11/19/2009 19:44'!
markDirty

	updatedOn := TimeStamp now.

	super markDirty.! !

!ARPersistentModel methodsFor: 'reactions' stamp: 'sebastian.sastre 11/19/2009 19:53'!
onAboutToDelete
	"The receiver is about to be deleted."
	
! !

!ARPersistentModel methodsFor: 'reactions' stamp: 'sebastiansastre 6/4/2011 19:11'!
onAboutToSave
	"The receiver is about to be saved."

	super onAboutToSave.

	updatedOn := TimeStamp now! !

!ARPersistentModel methodsFor: 'reactions' stamp: 'sebastian.sastre 3/10/2011 10:13'!
onAfterDelete
	"The receiver was just deleted."
	
! !

!ARPersistentModel methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 19:42'!
reverseSecondsCreatedOn
	^ self secondsCreatedOn negated! !

!ARPersistentModel methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 19:42'!
reverseSecondsUpdatedOn
	^ self secondsUpdatedOn negated! !

!ARPersistentModel methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 19:42'!
secondsCreatedOn
	^createdOn asSeconds ! !

!ARPersistentModel methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 19:42'!
secondsUpdatedOn
	^self updatedOn asSeconds! !

!ARPersistentModel methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 19:31'!
updatedOn
	^ updatedOn! !

!ARPersistentModel methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 19:40'!
updatedOn: anObject
	^ updatedOn := anObject! !

ARPersistentModel subclass: #ARPersonMock
	instanceVariableNames: 'name dateOfBirth description father mother children buddy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Example'!

!ARPersonMock class methodsFor: 'as yet unclassified' stamp: 'seb 9/3/2011 12:50'!
testPerson
	^ (self new)
		name: 'mzqfmmv zptnhli';
		dateOfBirth: Date today - 12000 atRandom days;
		description: 'Rxvalpv tdpqkvv ikjcscw bigjmeb vukntxj qxshdtt wkczzio bqpqacu whluhqv.  Lqlwajh!!  Qknwpfd peeskqc oarthtv pfblwjj yndlxks.  .  Ucnlocq iuiiluk txkkzmh dshkhmq uokcwqy oiktkxl awxmpep rkkcmis xcgyoeg jodtjaf ntwijzp?  Bqpjsiq qthzwtx xcnqbag ausurqp xatoqar jytguon?  Bpigdij jgijxep bjgttpr kklsfaw rdoikwz!!  Hzdbjxc?  Ehwflyy qfcqntk mlhcjql ecbtrtk egxldpk rypalvw ndngbdl jhhdxts zrfjtaz gccrnni dwuqxwm nmifzcq sxkgkxh.  Kwtzsjv ghvjqqm lfgibdt rufcowp kavabmi bpveqsr shjzsft xqaxivy fjqydns ryyggif.

Jshooyy jsyojlv miusnaj onpdiss hypkzzd qaunpqd rrvgnws ekiswyv?  Glhxkkk?  Pzyjuot raaefbv cnxmbsd?  Jabaniw uirjkcm jocxnhd!!  Erwblox qikxfez.

Ezswaga cnrzhqi jrdikos kkcncxd eoewflm ylzilve viumpqm uiyqhvv azrezhc jjeesfd rfdgsgg wfmxfye iisisju vhjourb.

Yicqnqd zhaioos npnjjsc hxfwlfr ozynyjs!!  Qavvubu uodeedp?  Fjdgbxu ryswazg ifumqpv jtlifnd?  Blciucl kypalot?  Dstnwqj kwzpxrk bgewcyp triddbr acvrkey pwgcleb sqtxajv svzpnwk unnlmtg oqnbpsq svtnxdo!!

Ywodawk uikhwek kyqivqj iojkwiw.  Bjcsnpu jtfpmqd!!  Crwlngk qeeeuci nnakbai hncfdkk kbhehju ttdsdcf zqulfpj pmvtmfa xrusxuy wbamfee opzjdia hcrfdnz tiasrqz wkwidbg?  Tzlwcst twaeuwm tfhreal igamlby saekozk tjnxohl ogcdhva fckbaii utqvhjg?  Nrynpjl dniycpm kkvytuu bnuxrev zrvbcph.  Awfbhgg xtyffnz lwjkhdo evmvogt vqfqppp chdtxcr ktiwujg vcgqoya thgkhac ncagoxy unuukan gntyowz obmiwmf!!  Okwjujz tfghbxi jupstni xbzpgau zlxeblm!!  Llxuwqc gupeurb ltwyzzj xxanyln qgrtigr?

Vudeenm hnfwxay rtaacau ymmyxbi lpcbamg ifopjuw guhctxx radytrh yubjcjc lnjucta qinzmlf.  Cnunlvp gdhcgrq oxeojsd?  Bbkjtne kzlvdso xtmgqhj pfyroxl brmqhkt gvqnftz aupxtsg.

Krzxaye!!  Clhbjoq ubxtlsc jzhggvl updbzxe.

Vhtetst ruwpukn jdttpba?  Ctwactb ljoeiqa pubapwc cioporv?

Talkecn dbddhuw rxfybrf iwdxfdo spdbdyk vhxbgrm elqngon dytngbm knqxaqv zrltvgh snjkzig rzypaly ekabqcg.  Xxxmegt otxzkhh vxpyjkn pnhgpfx qsnrhjo ziftvuu ivvfacz nhdajef ezksurr gvfepxx.  Lqdldbo wegkzlk qxgqpux hfbtydd hbqfhea gxuksjm hvunuwa?  Lxqkddp bccukhy odljtnp nouapus dcqpkqz lgpxpcs fmaehsc cgybese gujcbqz icjzamy tvmvwlf dvtkpng!!  !!  Okqxkhh dfuulea?  Ratgmpz?  Pfmgfbk qhggllc.  Balobww?

Wqoyfli xlxenvi ngtdshb bhpmyhd.  Kxsidry wijdkem oyuutvw msunbhq vaektbt yfapmwl yjrwmxi xgzfpbk hxcmydy jgaybjh magjtcf.  Koxatqi ivvultq yptadcp ygqutkn dizybis nvsvfhx tdhcqfy ihvqsvs paakztf uzokxhk xlcfhbr wejdutf wywzrqs tnlswyo?  Glawqav nvndvdf cbqzhce iuygwer twyxquv pjklwnv tbnazjw vuvxvgi?

Rpyaiqq rnvfyzk fgbsbjc zhftunh uvmxaov vgvwzwq fpuurxc!!  Zemojud eadzsas wzfdcuv vxvulbc ahtiijq yclzbvf rspxhrc lymmwja!!  Mfmgrrb!!  Cyfuuvp iabqntx volbiij ebjihqw ypxvukv wnghvwo fhqwyuh tlmeyrl oxcmotx uemizkc xvbgzqh xsdqtcj xfbzrgg?  Jmtauxu!!  Kbigzux slwjszd manystg usxbiya kfgdygh wrclmih chqtyew orkrazt pekdcrh oexrivu bnbrerh?

Xbkcyvl xivutzr iitbwnm eouuiux iwaansa oxzygwa sosqxvi jkinurb bedwzss?  Uspwzru.  Eogmjki raowczb!!  ?  Lfzhatu ldnvhfn rzfqfcl yzhbxpo evahsye eibcwlx ygknqms terepdi acvezip lyqydat bksesji.  Sncraae sropzde xdhuuuf?  ?

Vtcevnw dqwwzqe kqmakan wboltvp axuhsio eddpqvk hmaqsst kbqyeqg qjnlkph zjrrcdr glyqfak!!  Ciofqeq qzfzdxn aumczcs yxzdrqu hxnlhmo ihnowav nqggjoi dfdxzqr.

Wiedtvt jpikavx hkxvzqi eyobbcb hxhktut qpgasux bnwnzhp xaceikr comlnmg jkdzrhn hctccgm zcjgqbc llklclt!!

Bgyypck fvgaauj bzetcjo rqixcnx ggwuzax cnknyiv vbtnkxj zkmenql ynxyqpp vqnvihq qydzixk jvsbxhw qnbphdl uddayfe!!  Glalnun gbtcspy kxnzokz ecdpgpr kcmvcdf ngczhkp oovhcik yvlzbgl?';
		yourself! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
buddy
	^ buddy! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
buddy: anObject
	buddy := anObject! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
children
	^ children! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
children: anObject
	children := anObject! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
dateOfBirth
	^ dateOfBirth! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
dateOfBirth: anObject
	dateOfBirth := anObject! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
description
	^ description! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
description: anObject
	description := anObject! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
father
	^ father! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
father: anObject
	father := anObject! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
mother
	^ mother! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
mother: anObject
	mother := anObject! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
name
	^ name! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
name: anObject
	name := anObject! !

!ARPersonMock methodsFor: 'accessing' stamp: 'seb 9/3/2011 12:50'!
refreshed
	^ self class atId: id! !

Object subclass: #ARCache
	instanceVariableNames: 'protection support'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Core'!

!ARCache methodsFor: 'actions' stamp: 'sebastian.sastre 11/25/2009 16:59'!
at: aTransaction named: aString
	"Answers the active records in the results stored 
	in the query results named aString related to a aTransaction
	or.. nil if none found."

	^ self protection critical:[|answer|
		answer := self support at: aTransaction ifAbsent:[nil].
		answer isNil 
			ifTrue:[answer]
			ifFalse:[answer results]]! !

!ARCache methodsFor: 'actions' stamp: 'sebastian.sastre 11/25/2009 16:49'!
at: aTransaction put: aQueryResults

	^ self protection critical:[
		self support at: aTransaction put: aQueryResults]! !

!ARCache methodsFor: 'actions' stamp: 'sebastian.sastre 11/26/2009 17:23'!
at: aTransaction put: someActiveRecords named: aString

	aTransaction when: #finishedTransaction: sendOnce: #onTransactionFinished: to: self.

	self at: aTransaction put: (ARQueryResults for: someActiveRecords named: aString).
	
	^ someActiveRecords ! !

!ARCache methodsFor: 'initialization' stamp: 'sebastian.sastre 11/25/2009 16:35'!
initializeProtection
	"Initializes the protection of the access to the cache.
	Note: cache is mostly to read so we can safely protect it 
	with a monitor so it allows reentrance."
	^ protection := Monitor new! !

!ARCache methodsFor: 'initialization' stamp: 'sebastian.sastre 11/25/2009 16:39'!
initializeSupport
	"Initializes the support of this cache.
	Note: currently it'll use an identity dictionary 
	which is not thread safe and needs the protection."
	^ support := IdentityDictionary new! !

!ARCache methodsFor: 'testing' stamp: 'sebastian.sastre 11/25/2009 23:05'!
isEmpty 

	^ self protection critical: [
		self support isEmpty ]! !

!ARCache methodsFor: 'reactions' stamp: 'sebastian.sastre 11/25/2009 17:29'!
onTransactionFinished: aTransaction

	self purgeAt: aTransaction ! !

!ARCache methodsFor: 'accessing' stamp: 'sebastian.sastre 11/25/2009 16:34'!
protection 

	^ protection ifNil:[self initializeProtection]! !

!ARCache methodsFor: 'actions' stamp: 'sebastian.sastre 11/25/2009 16:43'!
purge 
	"Purifies the support of this cache"
	
	self protection critical:[
		self support values do:[:aQueryResult|
			aQueryResult purge]].
	
	self initializeSupport 
	! !

!ARCache methodsFor: 'actions' stamp: 'sebastian.sastre 11/25/2009 17:00'!
purgeAt: aTransaction 
	"Purifies the support of this cache in the slot 
	that corresponds to aTransaction"
	
	self protection critical:[|queryResults|
		queryResults := self support at: aTransaction ifAbsent:[nil].
		queryResults isNil 
			ifTrue: [
				"Gentle fallback with a discrete warning"
				Transcript cr; show: 'warn: transaction not found when purging']
			ifFalse:[
				queryResults purge.
				self support removeKey: aTransaction]]! !

!ARCache methodsFor: 'actions' stamp: 'sebastian.sastre 11/25/2009 16:50'!
removeKey: aTransaction 

	^ self protection critical:[
		self support removeKey: aTransaction ifAbsent:[
			"Gentle fallback with a discrete warning"
			Transcript cr; show: 'warn: trying to remove from the cache a transaction that wasn''t there']]! !

!ARCache methodsFor: 'accessing' stamp: 'sebastian.sastre 11/25/2009 16:38'!
support 
	"Answers the object that holds the elements"
	^ support ifNil:[self initializeSupport]! !

Object subclass: #ARClassIndex
	instanceVariableNames: 'className selector allowsNil keySize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Indices'!

ARClassIndex subclass: #ARAttributeIndex
	instanceVariableNames: 'sortSelector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Indices'!

!ARAttributeIndex methodsFor: 'actions' stamp: 'sebastiansastre 9/30/2011 16:04'!
add: anAggregate attributes: someAttributes for: aStorage
	"Adds anAggregate relating it with someAttributes"

	| value set |

	anAggregate isNil ifTrue:[
		^ ARCantStoreNil signal].
	
	ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[		
		value := anAggregate perform: selector.

		(allowsNil not and:[
		value isNil]) ifTrue:[
			^ ARCantIndexNil signal].

		someAttributes at: selector put: value; markDirty.

		value isNil
			ifTrue: [set := self setForValue: self nilKey]
			ifFalse:[
				self nilKey = value ifTrue:[
					"A non nil value that is equal to the value used to index by the nil value.
					This is extreamely unlikely to happen but needs control.
					Note: the order, nil (first) = value (second) is because value could have
					a 'creative' implementation of = and nilKey won't (it's the string's =)"
					^ ARIndexCollisionWithNil signal].
				set := self setForValue: value].

		set add: anAggregate; markDirty]! !

!ARAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 23:13'!
at: anObject for: aStorage
	"Answers the aggregates found in this index related 
	to the 'value' anObject or an empty array if none found"

	(allowsNil not and:[
	anObject isNil]) ifTrue:[
		"This index don't allow nil values so it can't 
		answer consistently when is asked about a nil value"
		^ ARCantIndexNil signal].
	
	^ ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[|results|
		results := self readOnlySetForValue: anObject.
		results isNilOrEmpty 
			ifTrue: [#()]
			ifFalse: [results]]! !

!ARAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/13/2010 09:38'!
basicFrom: start to: end
	"Answers aggregates found in this index 
	related to the given interval (as occurrences)."

	| sets bunchSize answer |
	#deprecated.
	self assert: end >= start.
	
	bunchSize := (end - start) +1.
	sets := self indexingSupport fromPosition: 1 to: end.
	answer := TreeSet sortSelector: selector equalitySelector: #isIdenticalTo:.

	"1. Adds all as if starts were always 1 (one)"
	1 to: sets size do:[:i|
		answer size > end ifFalse:[|nextOne|
			nextOne := sets at:i.
			answer addAll: nextOne]].

	"2. remove spurious elements at the beggining of those results (from 1 to start)"
	start > 1 ifTrue:[
		(start - 1) timesRepeat: [
			answer remove: answer first]].

	"3. remove spurious elements at the end of those results"
	answer size > bunchSize ifTrue:[|spurious|
		spurious := answer size - bunchSize.
		spurious timesRepeat: [
			answer remove: answer last]].
	
	^ answer 
	! !

!ARAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 11/25/2009 13:38'!
first
	"Answers the first of the active records 
	that are currently maintained by this index.
	Or.. return nil"
	^ self indexingSupport isEmpty
		ifTrue: [nil]
		ifFalse: [self indexingSupport first value first]! !

!ARAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 14:37'!
first: anInteger
	"Answers the firsts anInteger aggregates
	that are currently maintained by this index.
	Or.. return an empty bunch"

	^ self indexingSupport isEmpty
		ifTrue: [#()]
		ifFalse: [(self indexingSupport first: anInteger) collect:[:e| e first]]! !

!ARAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 11/21/2009 13:38'!
from: start to: end do: aBlock for: aStorage
	"Answers the aggregates found in this index 
	related to the given interval."
	
	^ ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[
		self indexingSupport 
			from: start
			to: end 
			do:[:bunch| bunch do: aBlock]]! !

!ARAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/13/2010 15:10'!
from: start to: end for: aStorage
	"Answers the aggregates found in this index 
	related to the given interval (as occurrences)."
	
	"This design wasn't considering to handle occurrences so his implementation
	isn't efficient. It will do its work for small ranges though, specially if it starts 
	near the beggining of the index."
	^ ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[
		start <= end 
			ifTrue:[(ARIndexIterator on: self) from: start to: end]
			ifFalse:[(ARReverseIndexIterator on: self) from: start to: end]]! !

!ARAttributeIndex methodsFor: 'testing' stamp: 'sebastian.sastre 4/24/2010 14:00'!
hasSortSelector

	^ sortSelector notNil ! !

!ARAttributeIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 11/20/2009 12:16'!
indexName

	^ className asString,'-', selector asString! !

!ARAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 14:42'!
last
	"Answers the last of the active records 
	that are currently maintained by this index.
	Or.. return nil"
	^ self indexingSupport isEmpty
		ifTrue: [nil]
		ifFalse: [self indexingSupport last value first]! !

!ARAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 14:43'!
last: anInteger
	"Answers the lasts anInteger aggregates
	that are currently maintained by this index.
	Or.. return an empty bunch"

	^ self indexingSupport isEmpty
		ifTrue: [#()]
		ifFalse: [(self indexingSupport last: anInteger) collect:[:e| e first]]! !

!ARAttributeIndex methodsFor: 'private' stamp: 'sebastian.sastre 3/6/2011 14:19'!
makeSetFor: anAggregateOrValue 
	"Answers a new persisted set to store 
	the (potentially big quantity of) indexed aggregates."

	| set |	
		
	(self hasSortSelector not and:[
	anAggregateOrValue respondsTo: #id])
		ifTrue:[set := ARTreeSet sortSelector: #id equalitySelector: #=]
		ifFalse:[
			self hasSortSelector not ifTrue:[
				set := ARTreeSet sortSelector: #hash equalitySelector: #isIdenticalTo:]].
		

	"Note (sas): I can't tell right now if the values got using 
	sortSelector are persistently updateables so maybe is better 
	to use it only with immutable values (like creation date time)"
	self hasSortSelector ifTrue: [
		self assert: sortSelector isSymbol.
		set := ARTreeSet sortSelector: sortSelector equalitySelector: #=].

	^ set makePersistent! !

!ARAttributeIndex methodsFor: 'private' stamp: 'sebastian.sastre 4/26/2010 18:38'!
readOnlySetForValue: anAggregateOrValue
	"Answers the (potentially big) set of objects related to
	anAggregateOrValue regarding to this index or nil.
	Note: it won't create the set if it does not exists.'"

	^ self setOrNilForValue: anAggregateOrValue! !

!ARAttributeIndex methodsFor: 'actions' stamp: 'sebastiansastre 10/1/2011 06:42'!
remove: anAggregate attributes: someAttributes for: aStorage
	"Removes anAggregate un-relating it to someAttributes for aStorage."

	| value set oldAttribute setKey |
	ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[
		oldAttribute := someAttributes at: selector ifAbsent:[nil].

		"even if it doesn't allow nil, why complaining on a remove, right? 
		(oldAttribute isNil and:[
		allowsNil not]) ifTrue:[
			^ ARCantIndexNil signal]."

		(allowsNil and:[
		oldAttribute isNil])
			ifTrue: [setKey := self nilKey]
			ifFalse: [setKey := oldAttribute].
		set := self setOrNilForValue: setKey.
		set isNil ifFalse:[
			set remove: anAggregate; markDirty.
			set isEmpty ifTrue:[
				(oldAttribute isNil and:[
				allowsNil]) 
					ifTrue:[setKey := self nilKey]
					ifFalse:[setKey := oldAttribute].
				self indexingSupport
					removeKey: setKey;
					markDirty]]]! !

!ARAttributeIndex methodsFor: 'private' stamp: 'sebastian.sastre 4/24/2010 14:29'!
setForValue: anAggregateOrValue
	"Answers the (potentially big) set of objects related to
	anObject regarding to this index."

	^ self indexingSupport 
		at: anAggregateOrValue 
		ifAbsent:[|set|
			set := self makeSetFor: anAggregateOrValue.
			self indexingSupport 
				at: anAggregateOrValue put: set; 
				markDirty;
				yourself.
			set]! !

!ARAttributeIndex methodsFor: 'private' stamp: 'sebastian.sastre 4/26/2010 18:42'!
setOrNilForValue: anAggregateOrValue
	"Answers the (potentially big) set of objects related to
	anAggregateOrValue regarding to this index or nil if there isn't one."

	^ self indexingSupport 
		at: anAggregateOrValue 
		ifAbsent:[nil]! !

!ARAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 11/25/2009 13:29'!
size
	"Answers the quantity of active records are currently maintained by this index"
	^ self indexingSupport size! !

!ARAttributeIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 4/24/2010 13:43'!
sortSelector
	^ sortSelector ! !

!ARAttributeIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 4/24/2010 13:42'!
sortSelector: anObject
	sortSelector := anObject! !

ARAttributeIndex subclass: #ARManyAttributesIndex
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Indices'!

!ARManyAttributesIndex methodsFor: 'actions' stamp: 'sebastian.sastre 4/24/2010 12:07'!
add: anAggregate attributes: someAttributes for: aStorage
	"Adds anAggregate relating it with someAttributes"

	anAggregate isNil ifTrue:[
		^ ARCantStoreNil signal].
	
	ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[ |indexable set|

		indexable := anAggregate perform: selector.

		(allowsNil not and:[
		indexable isNil]) ifTrue:[
			^ ARCantIndexNil signal].

		((indexable notNil and:[indexable isMany not]) or:[
		(indexable isNil and:[allowsNil not])]) ifTrue:[
			"This index mantains (so expects) an attribute that answers many objects.
			Also nil (if allowed)"
			^ ARCantStoreMany signal].

		someAttributes at: selector put: indexable; markDirty.

		indexable isNil
			ifTrue: [
				set := self setForValue: self nilKey.
				set add: anAggregate; markDirty]
			ifFalse:[|indexableValue|
				indexable do:[:element|
					element = self nilKey ifTrue:[
					"A non nil element that is equal to the value used to index by the nil value.
					This is extreamely unlikely to happen but needs control"
					^ ARIndexCollisionWithNil signal].
				set := self setForValue: element.
				set add: anAggregate; markDirty]].
		
		"Currently we don't have an efficient way to control uniqueness of combinations
		(isUnique and:[set size > 1]) ifTrue:[
			ARKeyAlreadyStored signal]"]! !

!ARManyAttributesIndex methodsFor: 'queries' stamp: 'sebastian.sastre 11/26/2009 15:22'!
at: anObject for: aStorage
	"This kind of index refers to many values.
	It does not support groups of values treated as values"
	
	self shouldNotImplement ! !

!ARManyAttributesIndex methodsFor: 'queries' stamp: 'sebastian.sastre 11/26/2009 15:57'!
includes: anObject for: aStorage
	"Answers the active records found in this index related 
	to the 'value' anObject or an empty array if none found"

	(allowsNil not and:[
	anObject isNil]) ifTrue:[
		"This index don't allow nil values so it can't 
		answer consistently when is asked about a nil value"
		^ ARCantIndexNil signal].
	
	^ ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[|results|
		results := self readOnlySetForValue: anObject.
		results isNil
			ifTrue:[#()]
			ifFalse: [results]]! !

!ARManyAttributesIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 5/11/2010 23:46'!
indexName

	^ className asString,'-', selector asString! !

!ARManyAttributesIndex methodsFor: 'testing' stamp: 'sebastian.sastre 11/26/2009 15:19'!
isForMany

	^ true! !

!ARManyAttributesIndex methodsFor: 'actions' stamp: 'seb 9/3/2011 23:15'!
remove: anAggregate attributes: someAttributes for: aStorage
	"Removes anAggregate un-relating it to someAttributes for aStorage."

	| value set oldMany setKey |

	ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[
		oldMany := someAttributes at: selector  ifAbsent:[nil].

		(oldMany isNil and:[
		allowsNil not]) ifTrue:[
			^ ARCantIndexNil signal].

		oldMany isNil 
			ifTrue: [
				set := self setOrNilForValue: self nilKey.
				set isNil ifFalse:[
					self indexingSupport 
						removeKey: self nilKey;
						markDirty]]
			ifFalse: [
				oldMany do:[:element|
					element isNil ifTrue:[
						"It has no sense to allow mix nil in the attribute itself toghether with 
						some element in nil in the middle of many."
						^ ARCantIndexNil signal].
					set := self setOrNilForValue: element.
					set isNil ifFalse:[
						set remove: anAggregate; markDirty.
						set isEmpty ifTrue:[
						self indexingSupport
							removeKey: element;
							markDirty]]]]]! !

!ARManyAttributesIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 5/11/2010 23:46'!
sortSelector
	^ sortSelector ! !

!ARManyAttributesIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 5/11/2010 23:46'!
sortSelector: anObject
	sortSelector := anObject! !

!ARClassIndex class methodsFor: 'accessing' stamp: 'sebastiansastre 11/27/2011 12:27'!
defaultKeySize
	"Answers the value of the size of the key for
	this index.
	Note: It's small to reward efficiency but index configurators
	should set their own values carfully."

	^ 5! !

!ARClassIndex class methodsFor: 'accessing' stamp: 'sebastiansastre 8/18/2011 15:34'!
nilKey
	"Answers a value where nil is mapped on the indexing support.
	Note: has sense only if this index allows nil."
	^ OmniBase nilKey ! !

!ARClassIndex methodsFor: 'testing' stamp: 'sebastian.sastre 4/6/2010 17:14'!
= anIndex
	"Answers true if this index stores the same as anIndex"
	^ self aggregateClassName = anIndex aggregateClassName and:[
	self selector = anIndex selector]! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 23:19'!
aggregateClass
	"Answers the class ob objects that this index stores"
	
	^ Smalltalk at: className! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 23:19'!
aggregateClass: aClass
	"Stores the name of the class that exists in this image
	In future who knows..."

	"Applicability: renaming objects persisted in the storage 
	is not that rare and migrations use to need this kind of help"
	className := aClass name! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 23:13'!
aggregateClassName

	^ className! !

!ARClassIndex methodsFor: 'actions' stamp: 'sebastian.sastre 11/23/2009 13:20'!
allowNil

	allowsNil := true! !

!ARClassIndex methodsFor: 'testing' stamp: 'sebastian.sastre 11/23/2009 13:21'!
allowsNil
	"Answers true if this index can store nil as indexed value per active record"
	
	^ allowsNil ! !

!ARClassIndex methodsFor: 'queries' stamp: 'sebastian.sastre 11/20/2009 14:36'!
at: anObject for: aStorage
	"Answers the aggregates found in this index related 
	to the 'value' anObject"
	
	^ self subclassResponsibility ! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 11/20/2009 14:14'!
currentTransaction 

	^ OmniBase currentTransaction ! !

!ARClassIndex methodsFor: 'actions' stamp: 'sebastian.sastre 11/23/2009 13:22'!
dontAllowNil

	allowsNil := false! !

!ARClassIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 22:17'!
from: start to: end for: aStorage
	"Answers the aggregates found in this index 
	related to the given interval."
	
	^ ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[
		self indexingSupport fromPosition: start to: end]! !

!ARClassIndex methodsFor: 'testing' stamp: 'sebastian.sastre 11/19/2009 21:10'!
hasClass
	"Answers true if this index finds that className 
	is the name of a class that exists in this image."

	"Applicability: renaming objects persisted in the storage 
	is not that rare and migrations use to need this kind of checks"
	^ Smalltalk includesKey: className
	! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 11/20/2009 14:14'!
indexName

	^ self class name! !

!ARClassIndex methodsFor: 'testing' stamp: 'sebastian.sastre 11/26/2009 17:14'!
indexesClass: aClass
	"Answers true if this index actualli do index objects of aClass.
	Note: Aggregates indices need to be strict. 
	For indices there is no heritage."

	^ (Smalltalk includesKey: className) and:[
	aClass = self aggregateClass]
! !

!ARClassIndex methodsFor: 'testing' stamp: 'sebastian.sastre 4/25/2010 11:56'!
indexesClass: aClass selector: aSymbol

	^ (self indexesClass: aClass) and:[
	self indexesSelector: aSymbol]! !

!ARClassIndex methodsFor: 'testing' stamp: 'sebastian.sastre 4/25/2010 11:56'!
indexesSelector: aSymbol

	^ selector = aSymbol ! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 11/20/2009 14:14'!
indexingSupport
	"Answers the indexing support of this index"

	^ self currentTransaction root at: self indexingSupportKey! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 11/20/2009 14:16'!
indexingSupportKey
	"Answers the key where the indexing support 
	of this index is located in the repository"

	^ className,'>>',selector ! !

!ARClassIndex methodsFor: 'initializing' stamp: 'sebastian.sastre 5/24/2010 09:36'!
initialize

	super initialize.
	
	"Permisive by default"
	allowsNil := true.! !

!ARClassIndex methodsFor: 'initializing' stamp: 'sebastiansastre 9/25/2011 13:47'!
initializeKeySize
	
	^ keySize := self class defaultKeySize
	! !

!ARClassIndex methodsFor: 'testing' stamp: 'sebastian.sastre 11/26/2009 15:19'!
isForMany

	^ false! !

!ARClassIndex methodsFor: 'testing' stamp: 'sebastian.sastre 5/11/2010 16:26'!
isUnique
	"Answers true if this index maintains only one value per aggregate"
	
	^ false ! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastiansastre 9/25/2011 13:30'!
keySize
	
	^ keySize ifNil:[self initializeKeySize]
	! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastiansastre 9/25/2011 13:33'!
keySize: anInteger
	
	anInteger < 5 ifTrue:[self error: 'Key is too small'].
	anInteger > 255 ifTrue:[self error: 'Key is too big'].
	
	keySize := anInteger
	! !

!ARClassIndex methodsFor: 'testing' stamp: 'seb 9/3/2011 23:12'!
matches: someAttributes for: anAggregate
	"Answers true if, regarding to this index, 
	anAggregate matches someAttributes of false otherwise."
	
	^ (anAggregate perform: selector) = (someAttributes at: selector ifAbsent:[])! !

!ARClassIndex methodsFor: 'private' stamp: 'sebastiansastre 9/25/2011 13:30'!
newIndexingSupport
	"Answers a new instance of something to aid indexing"
		
	^ OmniBase newBTreeDictionary: self keySize
	
	"^ OmniBase newBTreeIdentityDictionary makePersistent "
	"^ ARBTree order: 32"! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastiansastre 8/18/2011 15:31'!
nilKey
	"Answers a value where nil is mapped on the indexing support.
	Note: has sense only if this index allows nil."
	^ self class nilKey ! !

!ARClassIndex methodsFor: 'private' stamp: 'sebastian.sastre 11/20/2009 16:08'!
printOn: aStream

	super printOn: aStream.

	aStream nextPutAll: ' (',self aggregateClassName,'-',selector asString,')'! !

!ARClassIndex methodsFor: 'actions' stamp: 'sebastian.sastre 11/20/2009 14:15'!
removeIndexingSupport
	"Removes the indexing support (if any)"
	
	^ self currentTransaction root 
		removeKey: self indexingSupportKey 
		ifAbsent: nil! !

!ARClassIndex methodsFor: 'actions' stamp: 'sebastian.sastre 11/20/2009 14:15'!
resetIndexingSupport
	"Installs a brand new indexing support"
	
	^ self currentTransaction root 
		at: self indexingSupportKey 
		put: self newIndexingSupport! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 11/20/2009 14:06'!
selector
	^ selector! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 11/20/2009 14:06'!
selector: anObject
	selector := anObject! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 11/20/2009 14:15'!
size

	^ self indexingSupport size! !

!ARClassIndex methodsFor: 'actions' stamp: 'sebastiansastre 10/17/2011 23:29'!
update: anAggregate attributes: someAttributes for: aStorage
	"Update (if needs to) the necessary indices for anAggregate"

	ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[
		self remove: anAggregate attributes: someAttributes for: aStorage.
		self add: anAggregate attributes:  someAttributes for: aStorage]! !

!ARClassIndex methodsFor: 'accessing' stamp: 'sebastiansastre 1/10/2012 11:57'!
values

	^ self indexingSupport values! !

ARClassIndex subclass: #ARUniqueAttributeIndex
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Indices'!

!ARUniqueAttributeIndex methodsFor: 'actions' stamp: 'sebastiansastre 10/2/2011 23:54'!
add: anAggregate attributes: someAttributes for: aStorage
	"Adds anAggregate relating it with someAttributes"

	| value set |

	anAggregate isNil ifTrue:[
		^ ARCantStoreNil signal].
	
	ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[
		value := anAggregate perform: selector.

		(allowsNil not and:[
		value isNil]) ifTrue:[
			^ ARCantIndexNil signalFor: anAggregate].

		someAttributes at: selector put: value; markDirty.
	
		value isNil
			ifTrue: ["self error: 'this should never happen'"
				self indexingSupport 
					at: self nilKey
					put: anAggregate;
					markDirty;
					yourself]
			ifFalse:[|found|
				found := self indexingSupport at: value.
				(found notNil and:[
				(found isIdenticalTo: anAggregate) not]) ifTrue:[
					ARKeyAlreadyStored signalFor: anAggregate].
				value = self nilKey ifTrue:[
					"This exceptionis is just bad luck... if it happens it means 
					you tried to use a value that is reserved for this index.
					A non nil value that is equal to the value used to index by the nil value.
					This is extreamely unlikely to happen but needs control"
					^ ARIndexCollisionWithNil signalFor: anAggregate].
				self indexingSupport 
					at: value 
					put: anAggregate;
					markDirty;
					yourself]]! !

!ARUniqueAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 12/23/2010 07:57'!
at: anObject for: aStorage
	"Answers the active records found in this index related 
	to the 'value' anObject or an empty array if none found"

	(allowsNil not and:[
	anObject isNil]) ifTrue:[
		"This index don't allow nil values so it can't 
		answer consistently when is asked about a nil value"
		^ ARCantIndexNil signal].
	
	^ ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[|result|
		self indexingSupport 
			at: anObject 
			ifAbsent:[nil]]! !

!ARUniqueAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 17:22'!
first
	"Answers the first of the active records 
	that are currently maintained by this index.
	Or.. return nil"
	^ self indexingSupport isEmpty
		ifTrue: [nil]
		ifFalse: [self indexingSupport first value]! !

!ARUniqueAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 17:22'!
first: anInteger
	"Answers the firsts anInteger aggregates
	that are currently maintained by this index.
	Or.. return an empty bunch"

	^ self indexingSupport isEmpty
		ifTrue: [#()]
		ifFalse: [self indexingSupport first: anInteger]! !

!ARUniqueAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 16:24'!
from: start to: end do: aBlock for: aStorage
	"Answers the aggregates found in this index 
	related to the given interval."
	
	^ ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[
		self indexingSupport 
			from: start
			to: end 
			do:[:bunch| bunch do: aBlock]]! !

!ARUniqueAttributeIndex methodsFor: 'queries' stamp: 'sebastiansastre 5/9/2011 21:12'!
from: start to: end for: aStorage
	"Answers the aggregates found in this index 
	related to the given interval."
	
	^ ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[
		self indexingSupport from: start to: end]! !

!ARUniqueAttributeIndex methodsFor: 'queries' stamp: 'sebastiansastre 5/10/2011 10:60'!
fromPosition: start to: end for: aStorage
	"Answers the aggregates found in this index 
	related to the given interval."
	
	^ ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[
		self indexingSupport fromPosition: start to: end]! !

!ARUniqueAttributeIndex methodsFor: 'testing' stamp: 'sebastian.sastre 5/24/2010 09:38'!
hasSortSelector

	^ false! !

!ARUniqueAttributeIndex methodsFor: 'accessing' stamp: 'sebastian.sastre 5/11/2010 16:24'!
indexName

	^ className asString,'-', selector asString! !

!ARUniqueAttributeIndex methodsFor: 'testing' stamp: 'sebastian.sastre 5/11/2010 16:26'!
isUnique
	"Answers true if this index maintains only one value per aggregate"
	
	^ true ! !

!ARUniqueAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 17:22'!
last
	"Answers the last of the active records 
	that are currently maintained by this index.
	Or.. return nil"
	^ self indexingSupport isEmpty
		ifTrue: [nil]
		ifFalse: [self indexingSupport last value]! !

!ARUniqueAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 17:23'!
last: anInteger
	"Answers the lasts anInteger aggregates
	that are currently maintained by this index.
	Or.. return an empty bunch"

	^ self indexingSupport isEmpty
		ifTrue: [#()]
		ifFalse: [self indexingSupport last: anInteger]! !

!ARUniqueAttributeIndex methodsFor: 'actions' stamp: 'sebastiansastre 10/1/2011 06:42'!
remove: anAggregate attributes: someAttributes for: aStorage
	"Removes anAggregate un-relating it to someAttributes for aStorage."

	| toRemove key |

	ARCurrentContainer use: (aStorage containerNameForAttributeIndex: self) during:[
		toRemove := someAttributes at: selector ifAbsent:[nil].

		"even if it doesn't allow nil, why complaining on a remove, right? 
		(toRemove isNil and:[
		allowsNil not]) ifTrue:[
			^ ARCantIndexNil signal]."

		(allowsNil and:[
		toRemove isNil])
			ifTrue: [key := self nilKey]
			ifFalse: [key := toRemove].

		self indexingSupport 
			removeKey: key;
			markDirty;
			yourself]! !

!ARUniqueAttributeIndex methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 16:24'!
size
	"Answers the quantity of active records are currently maintained by this index"
	^ self indexingSupport size! !

Object subclass: #ARContainerPolicy
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Core'!

!ARContainerPolicy commentStamp: 'sebastian.sastre 11/19/2009 20:35' prior: 0!
ARContainerPolicy is the abstract (parent) class of the policies for deciding which container should be used to persist a new object in the repository.

There are a lot of possible ways to do this so see subclasses for a more concrete behavior.

Beware: during the life-span of a repository is not recomended to change the policy. It's probably less error prone to re-incarnate the whole repository in a new repository with a new storage that uses the new policy.!

!ARContainerPolicy methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 22:18'!
containerFor: anAggregate in: aStorage
	"Answers the container to be used for anAggregate"
	
	^ self subclassResponsibility ! !

!ARContainerPolicy methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 22:17'!
containerNameFor: anAggregate in: aStorage
	"Answers the name of the container to be used for anAggregate"

	^ self subclassResponsibility ! !

!ARContainerPolicy methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 23:28'!
containerNameForAttributeIndex: anIndex in: aStorage
	"Answers the name of the container to be used for the indexed attributes of the aggregates indexed by anIndex"

	self subclassResponsibility 
	! !

!ARContainerPolicy methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 23:11'!
containerNameForIndex: anIndex in: aStorage

	self subclassResponsibility 
	! !

!ARContainerPolicy methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 22:18'!
indexedAttributesContainerNameFor: anAggregate in: aStorage
	"Answers the name of the container to be used for the indexed attributes of anAggregate"

	self subclassResponsibility 
	! !

ARContainerPolicy subclass: #ARDefaultContainer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Core'!

!ARDefaultContainer commentStamp: 'sebastian.sastre 11/19/2009 20:37' prior: 0!
An ARDefaultContainer is the policy of deciding to send every new persisted object to the default repository.

If you are using a repository for development this is a no-brainer. In the other hand, for big respositories or production storages, probably other policies are a wiser option.!

!ARDefaultContainer methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 22:23'!
containerFor: anAggregate in: aStorage
	"Answers the container to be used for anAggregate"

	^ [:container|
		container notNil
			ifTrue:[container containerNamed: self defaultContainerName ifAbsent: [
						self error: 'Something unexpected (and basic) is not right about this repository']]
			ifFalse:[self error: 'Can''t decide which container to use :(']] value: ARCurrentClient value! !

!ARDefaultContainer methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 22:22'!
containerNameFor: anAggregate in: aStorage
	"Answers the name of the container to be used for anAggregate"

	^ self defaultContainerName ! !

!ARDefaultContainer methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 23:29'!
containerNameForAttributeIndex: anIndex in: aStorage
	"Answers the name of the container to be used for the indexed attributes of the aggregates indexed by anIndex"

	^ self defaultContainerName ! !

!ARDefaultContainer methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 23:11'!
containerNameForIndex: anIndex in: aStorage

	^ self defaultContainerName ! !

!ARDefaultContainer methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 22:23'!
defaultContainerName
	"Answers the name of the container that is the default for OmniBase"

	^ 'Default'! !

!ARDefaultContainer methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 22:23'!
indexedAttributesContainerNameFor: anAggregate in: aStorage
	"Answers the name of the container to be used for the indexed attributes of anAggregate"

	^ self defaultContainerName ! !

Object subclass: #ARODBPool
	instanceVariableNames: 'path clientsMutex busyClients idleClients'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Core'!

!ARODBPool commentStamp: 'sebastian.sastre 11/19/2009 20:31' prior: 0!
An ARODBPool is a resource to access the repository in an ACID fashion.
You'll be able to commit or evaluate.
It tries hard to be thread-safe.!

!ARODBPool class methodsFor: 'initializing' stamp: 'sebastian.sastre 11/23/2009 10:45'!
initialize

	Smalltalk addToShutDownList: self! !

!ARODBPool class methodsFor: 'actions' stamp: 'sebastiansastre 1/6/2013 19:42'!
onPath: aDirectoryName
	"Answers a new pool to work on aDirectoryName"

	^ self new
		initializeOnPath: aDirectoryName asReference;
		yourself! !

!ARODBPool class methodsFor: 'actions' stamp: 'sebastian.sastre 11/23/2009 10:44'!
resetAll

	self allSubInstancesDo: [:ea | ea reset]! !

!ARODBPool class methodsFor: 'actions' stamp: 'sebastian.sastre 11/23/2009 10:44'!
shutDown

	self resetAll! !

!ARODBPool methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 12:46'!
busyClients 
	"Answers the bunch of clients that are currently busy"
	^ busyClients ifNil:[self initializeBusyClients]! !

!ARODBPool methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 13:18'!
clientsMutex
	"Answers the mutex used to protect clients usage"
	^ clientsMutex ifNil: [self initializeClientsMutex]! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastiansastre 1/5/2013 22:22'!
commit: aBlock
	"Evaluates aBlock in an atomic 
	transactional environment that actually can write things"

	^ self commit: aBlock readOnly: false! !

!ARODBPool methodsFor: 'private' stamp: 'sebastian.sastre 11/19/2009 14:18'!
commit: aBlock readOnly: aBoolean
	"Evaluates aBlock commiting any change to the repository.
	Read only or not depends on aBoolean's value"

	^ self evaluate: aBlock readOnly: aBoolean commit: true! !

!ARODBPool methodsFor: 'private' stamp: 'sebastian.sastre 1/1/2011 12:44'!
createBackupOf: anOmniBase on: newDirectoryName
	"Starts an on-line backup process creating a (hot) backup copy
	of the receiver database." 
	
	anOmniBase createBackupOn: newDirectoryName asString! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastian.sastre 8/11/2010 14:19'!
createBackupOn: newDirectoryName
	"Starts an on-line backup process creating a (hot) backup copy
	of the receiver database." 
	
	self withClientDo:[:anOmniBase|
		anOmniBase createBackupOn: newDirectoryName asString]! !

!ARODBPool methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 20:09'!
currentClient
	"Answers the client used by the currently active process.
	
	This is here as convenience and it will make sense 
	to send it inside a transactional environment like this:

	aPool commit:[aPool currentClient doLowLevelStuff]"
	
	^ ARCurrentClient value! !

!ARODBPool methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 13:40'!
currentTransaction
	"Answers the transaction of the currently active process.
	
	This is here as convenience and it will make sense 
	to send it inside a transactional environment like this:

	aPool commit:[aPool currentTransaction doLowLevelStuff]"
	
	^ OmniBase currentTransaction ! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastiansastre 3/4/2012 14:04'!
destroy

	self reset.
	
	path deleteAll
	! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastian.sastre 1/7/2011 18:22'!
ensureExistence
	"Makes sure that an OmniBase repository 
	exists in path. It will only create a repo
	if the path does not (previously) exists"

	self isLocked ifTrue:[ ^self].
	
	(ODBFileStream directoryExists: path asString) ifFalse:[
		path asPath ensureDirectory.
		self lockOdb.
		[OmniBase createOn: path asString]
			ensure:[self unlockOdb]].
! !

!ARODBPool methodsFor: 'private' stamp: 'sebastian.sastre 11/19/2009 14:17'!
evaluate: aBlock readOnly: aBoolean
	"Evaluates aBlock without commiting any change to the repository.
	Read only or not depends on aBoolean's value"
	
	^ self evaluate: aBlock readOnly: aBoolean commit: false! !

!ARODBPool methodsFor: 'private' stamp: 'sebastian.sastre 2/15/2010 12:48'!
evaluate: aBlock readOnly: isReadOnly commit: hasToCommit
	"Evaluates aBlock commiting changes to the repository depending on hasToCommit.
	Read only or not depends on isReadOnly's value"

	^ self withClientDo:[:omnibase| |activeProcess currentTransaction previousTransaction|
		activeProcess := Processor activeProcess.
		currentTransaction := isReadOnly
						ifTrue: [omnibase newReadOnlyTransaction]
						ifFalse: [omnibase newTransaction].

		previousTransaction := OmniBase currentTransactionOrNil.

		OmniBase currentTransaction: currentTransaction.
		[hasToCommit
			ifTrue:[
				aBlock commitIn: currentTransaction]
			ifFalse:[
				aBlock evaluateIn: currentTransaction]] ensure:[
					OmniBase currentTransaction: previousTransaction]]! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastian.sastre 1/31/2011 13:52'!
exclusiveCommit: aBlock
	"Evaluates aBlock in an atomic 
	transactional environment that actually can write things"
	^ self exclusiveCommit: aBlock readOnly: false! !

!ARODBPool methodsFor: 'private' stamp: 'sebastian.sastre 1/31/2011 13:52'!
exclusiveCommit: aBlock readOnly: aBoolean
	"Evaluates aBlock commiting any change to the repository.
	Read only or not depends on aBoolean's value"

	^ self exclusiveEvaluate: aBlock readOnly: aBoolean commit: true! !

!ARODBPool methodsFor: 'private' stamp: 'sebastian.sastre 1/31/2011 13:51'!
exclusiveEvaluate: aBlock readOnly: aBoolean
	"Evaluates aBlock without commiting any change to the repository.
	Read only or not depends on aBoolean's value"
	
	^ self exclusiveEvaluate: aBlock readOnly: aBoolean commit: false! !

!ARODBPool methodsFor: 'private' stamp: 'sebastian.sastre 1/31/2011 13:51'!
exclusiveEvaluate: aBlock readOnly: isReadOnly commit: hasToCommit
	"Evaluates aBlock commiting changes to the repository depending on hasToCommit.
	Read only or not depends on isReadOnly's value"

	^ self withExclusiveClientDo: [:omnibase| |activeProcess currentTransaction previousTransaction|
		activeProcess := Processor activeProcess.
		currentTransaction := isReadOnly
						ifTrue: [omnibase newReadOnlyTransaction]
						ifFalse: [omnibase newTransaction].

		previousTransaction := OmniBase currentTransactionOrNil.

		OmniBase currentTransaction: currentTransaction.
		[hasToCommit
			ifTrue:[
				aBlock commitIn: currentTransaction]
			ifFalse:[
				aBlock evaluateIn: currentTransaction]] ensure:[
					OmniBase currentTransaction: previousTransaction]]! !

!ARODBPool methodsFor: 'testing' stamp: 'sebastiansastre 7/16/2011 15:09'!
exists
	"Answers true if this pool detects an existing repository"
	
	^ self hasClients or:[
	"The existence test is not currently superb, it's just a directory check"
	path asPath exists and:[
	self hasOmniBaseFiles and:[
	self hasRoot]]]! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastiansastre 1/5/2012 11:52'!
garbageCollect

	self triggerEvent: #aboutToGarbageCollect: with: self.
	
	"Free all the clients of this odb"
	self reset.
	"To do: broadcast a free clients request and wait rasonably."
	
	^ self withExclusiveClientDo:[:omnibase|
		self garbageCollectOn: omnibase]! !

!ARODBPool methodsFor: 'private' stamp: 'sebastiansastre 12/7/2011 16:00'!
garbageCollectOn: anOmniBase

	^ anOmniBase garbageCollect! !

!ARODBPool methodsFor: 'testing' stamp: 'sebastian.sastre 11/30/2009 13:56'!
hasBusyClients
	"Answers true if this pool has, at least one, busy client"
	^ self busyClients notEmpty! !

!ARODBPool methodsFor: 'testing' stamp: 'sebastian.sastre 11/30/2009 13:55'!
hasClients
	"Answers true if this pool has at least one client"
	^ self hasIdleClients or:[
	self hasBusyClients]! !

!ARODBPool methodsFor: 'testing' stamp: 'sebastian.sastre 11/19/2009 13:27'!
hasIdleClients
	"Answers true if this pool has, at least one, idle client"
	^ self idleClients notEmpty! !

!ARODBPool methodsFor: 'testing' stamp: 'sebastiansastre 4/20/2012 14:18'!
hasOmniBaseFiles 

	^ path exists and:[|files|
	files := path allFiles collect:[:f| f basename].
	#(
	'omnibase.log'
	'omnibase.cif'
	'omnibase.cid'
	'omnibase.odt'
	'omnibase.vcf'
	) allSatisfy: [:e|
		files includes: e ]]
! !

!ARODBPool methodsFor: 'testing' stamp: 'sebastiansastre 7/15/2011 16:04'!
hasRoot

	^ [self readOnly:[self root notNil]] on: Error do:[:x| ^ false]! !

!ARODBPool methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 12:46'!
idleClients 
	"Answers the bunch of clients that are currently idle"
	^ idleClients ifNil:[self initializeIdleClients]! !

!ARODBPool methodsFor: 'initialization' stamp: 'sebastian.sastre 11/19/2009 12:47'!
initializeBusyClients

	^ busyClients := IdentitySet new! !

!ARODBPool methodsFor: 'initialization' stamp: 'sebastian.sastre 11/19/2009 13:18'!
initializeClientsMutex

	^ clientsMutex := Semaphore forMutualExclusion! !

!ARODBPool methodsFor: 'initialization' stamp: 'sebastian.sastre 11/19/2009 12:47'!
initializeIdleClients

	^ idleClients := IdentitySet new! !

!ARODBPool methodsFor: 'initialization' stamp: 'sebastian.sastre 12/24/2010 11:41'!
initializeOnPath: aPath
	"Sets the path of the OmniBase repository 
	that this pool will provide access to"
	path := aPath ! !

!ARODBPool methodsFor: 'testing' stamp: 'sebastian.sastre 1/1/2011 12:17'!
isLocked
	"Answers true if this odb is locked (i.e. being garbage collected 
	or something that requires exclusive access).
	Note: this is implemented with a file presence/absence as lock so that state
	is known by any worker image trying to access it."
	
	^ self isLockedFile exists! !

!ARODBPool methodsFor: 'accessing' stamp: 'sebastian.sastre 1/1/2011 12:16'!
isLockedFile

	^ path / 'isLocked'! !

!ARODBPool methodsFor: 'testing' stamp: 'sebastian.sastre 1/1/2011 12:17'!
isUnderMaintenance
	"Answers true if this odb is under mainteance (i.e. being garbage collected 
	or something that requires exclusive access).
	Note: this is implemented with a file presence/absence as lock so that state
	is known by any worker image trying to access it."
	
	^ self isLocked ! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastiansastre 7/16/2011 17:40'!
lockOdb
	"Locks the odb."

	(self numberOfUsers > 1 or:[
	(self numberOfUsers = 1 and:[
	busyClients size = 1])]) ifTrue:[
		^ ARCantLockOdb signal].
	
	self isLockedFile ensureFile! !

!ARODBPool methodsFor: 'accessing' stamp: 'sebastian.sastre 1/1/2011 12:12'!
numberOfUsers
	"Answers how many users are connected to the odb."
	
	^ self withClientDo:[:anOmniBase|
		anOmniBase numberOfClients]! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastiansastre 4/20/2012 14:17'!
openOmniBase
	"Returns a new OmniBase instance 
	opened in the path of this pool.
	It will create a repository when none is found"

	path exists ifFalse:[
		path ensureDirectory].

	^ self hasOmniBaseFiles
		ifFalse:[OmniBase createOn: path asString]
		ifTrue: [OmniBase openOn: path asString].! !

!ARODBPool methodsFor: 'accessing' stamp: 'sebastian.sastre 1/7/2010 16:26'!
path 

	^ path! !

!ARODBPool methodsFor: 'private' stamp: 'sebastian.sastre 11/19/2009 14:02'!
printOn: aStream

	super printOn: aStream.

	aStream nextPutAll: '(',self hash asString,' on ',path asString,' ', self idleClients size asString,' idle, ',self busyClients size asString,' busy)'! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastiansastre 10/20/2012 12:37'!
purgeClients

	self clientsMutex critical:[
		self busyClients do:[:client|
			client isUsable ifFalse:[self busyClients remove: client ifAbsent:[nil]]].
					
		self idleClients do:[:client|
			client isUsable ifFalse:[self idleClients remove: client ifAbsent:[nil]]]].! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastiansastre 3/3/2012 12:10'!
purify
	"Makes a garbage collection and then evolves.
	It should leave the new odb with the cremme de la cremme"
	
	self triggerEvent: #aboutToPurify: with: self.
	
	"Free all the clients of this odb"
	self reset.
	"To do: broadcast a free clients request and wait rasonably."

	"The garbage collection before the hot backup is needed
	because sometimes there are residual versions that would
	throw an exception interrupting the backup."
	self withExclusiveClientDo:[:anOmniBase|
		self 
			reorganizeOn: anOmniBase;
			purifyOn: anOmniBase].

	self reset.

	self removePreviousGeneration.

	(path / 'lastPurifyOn') ensureFile writeStreamDo:[:str| 
		str nextPutAll: TimeStamp current asString]! !

!ARODBPool methodsFor: 'private' stamp: 'sebastiansastre 3/3/2012 12:11'!
purifyOn: anOmniBase
	"Returns the next generation of this odb.
	This is what this does:
	1. create a hot backup of the current odb in path.NG
	2. closes this odb.
	3. renames this odb to path.old
	4. renames path.NG to path."
	
	| nextGeneration |
	nextGeneration := path parent / (path basename,'.NG').
	nextGeneration exists ifTrue:[
		"Discarding a failed successor isn't key for evolution?"
		nextGeneration deleteAll].

	self createBackupOf: anOmniBase on: nextGeneration.
	path renameAs: (path basename,'.old').
	nextGeneration renameAs: path basename.
	! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 14:38'!
readOnly: aBlock
	"Evaluates aBlock in an atomic 
	transactional environment that can't write anything"
	^ self evaluate: aBlock readOnly: true! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastian.sastre 1/1/2011 12:46'!
removeOldOdb
	"Removes the old odb version of this odb (if present).
	Answers the size bytes that were cleaned up."
	
	| size oldPath |

	oldPath := path parent / (path basename,'.old').
	oldPath exists ifFalse:[
		^0].
	
	size := oldPath allEntries inject: 0 into:[:sum  :e| sum + e size].
	oldPath deleteAll.
		
	^ size 
! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastian.sastre 12/30/2010 11:42'!
removePreviousGeneration
	"Removes the old odb version of this odb (if present).
	Answers the size bytes that were cleaned up."
	
	^ self removeOldOdb! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastiansastre 1/5/2012 11:53'!
reorganize

	self triggerEvent: #aboutToReorganize: with: self.
	
	"Free all the clients of this odb"
	self reset.
	"To do: broadcast a free clients request and wait rasonably."
	
	^ self withExclusiveClientDo:[:omnibase|
		self reorganizeOn: omnibase]! !

!ARODBPool methodsFor: 'private' stamp: 'sebastiansastre 12/7/2011 16:00'!
reorganizeOn: anOmniBase

	^ anOmniBase reorganize! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastiansastre 10/20/2012 12:20'!
requestClient
	"Answers an OmniBase instance taken from the idle clients.
	It's a thread safe operation."
	| anOmniBase |

	self clientsMutex critical: [
		"Transcript cr; show: 'getting odb client p: ', Processor activeProcess hash asString."
		anOmniBase := self hasIdleClients
						ifTrue: [self idleClients anyOne]
						ifFalse: [self openOmniBase].
		self idleClients remove: anOmniBase ifAbsent:[nil].
		self busyClients add: anOmniBase].
	
	^ anOmniBase! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastian.sastre 1/9/2011 17:54'!
reset
	"Closes all the OmniBase open clients"

	self clientsMutex critical:[
		self busyClients isEmpty ifFalse: [
			ARResetingWithOnlineClients signal: 'There are clients still in the used list.  
Beware that closing busy clients is not threadsafe.'].
		
		self idleClients do: [:anOpenOmniBase | 
			anOpenOmniBase notNil ifTrue: [
				anOpenOmniBase close]].
		
		idleClients := nil.
		
		self busyClients do: [:anOpenOmniBase | 
			anOpenOmniBase notNil ifTrue: [
				anOpenOmniBase close]].
		
		busyClients := nil]! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 13:37'!
returnClient: anOmniBase
	"Makes anOmniBase to go from the busy ones 
	to the idle ones"

	self clientsMutex critical:[
		self busyClients remove: anOmniBase ifAbsent: [self returningUnusedClient].		
		self idleClients add: anOmniBase]! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastian.sastre 11/19/2009 13:33'!
returningUnusedClient

	Transcript cr; show: 'Warning: returning a client that is not in among the busy ones'! !

!ARODBPool methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 13:36'!
root
	"Answers the root of the repository.
	This is here as convenience and it will make sense 
	to send it inside a transactional environment like this:

	aPool commit:[(aPool root at: #stuff) singAndDance]"
	
	^ OmniBase root! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastiansastre 6/16/2011 15:48'!
storage 

	^ self currentTransaction root at: #storage! !

!ARODBPool methodsFor: 'actions' stamp: 'sebastian.sastre 1/1/2011 12:21'!
unlockOdb
	"Unlocks the odb."

	self isLockedFile delete! !

!ARODBPool methodsFor: 'private' stamp: 'sebastian.sastre 11/19/2009 20:10'!
withClientDo: aBlock
	"Evaluates aBlock using one of the 
	(previously) idle clients as parameter
	and with it accessible as AGCurrentClient value 
	during the whole evaluation"

	| client |

	[[client := self requestClient] 
		on: ODBLockNotification 
		do:[:notification|
			Transcript cr; show: 'lock notification'.
			notification resume]] 
				on: ODBStreamLockedNotification
				do:[:notification| 
					Transcript cr; show: 'stream lock notification'.
					notification resume].
				
	^ ARCurrentClient use: client during:[
		[aBlock value: client] ensure:[
			"Transcript cr; show: 'returning client c: ', omnibase asString."
			self returnClient: client]]
! !

!ARODBPool methodsFor: 'private' stamp: 'sebastiansastre 7/16/2011 17:35'!
withExclusiveClientDo: aBlock
	"Evaluates aBlock using one of the idle clients using it as parameter.
	It will be accessible during the whole evaluation using the 
	dynamic variable: AGCurrentClient value
	The evaluation is exclusive, so it will set the locked file"

	| client |

	[[client := self requestClient] 
		on: ODBLockNotification 
		do:[:notification|
			Transcript cr; show: 'lock notification'.
			notification resume]] 
				on: ODBStreamLockedNotification
				do:[:notification| 
					Transcript cr; show: 'stream lock notification'.
					notification resume].
				
	^ ARCurrentClient use: client during:[
		[
		self lockOdb.
		aBlock value: client] ensure:[
			"Transcript cr; show: 'returning client c: ', omnibase asString."
			self unlockOdb.
			self returnClient: client.
			self reset]]
! !

Object subclass: #ARPersistentIntrospector
	instanceVariableNames: 'introspectee'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Introspection'!

!ARPersistentIntrospector class methodsFor: 'categorize me' stamp: 'sas 10/16/2008 20:20'!
for: anObject

	^ self new 
		introspectee: anObject;
		yourself! !

!ARPersistentIntrospector methodsFor: 'accessing' stamp: 'sas 10/16/2008 20:20'!
introspectee
	^introspectee! !

!ARPersistentIntrospector methodsFor: 'accessing' stamp: 'sas 10/16/2008 20:20'!
introspectee: anObject
	introspectee := anObject! !

ARPersistentIntrospector subclass: #ARPersistentReader
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Introspection'!

ARPersistentReader subclass: #ARCollectionPersistentReader
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Introspection'!

!ARCollectionPersistentReader methodsFor: 'actions' stamp: 'sebastian.sastre 2/3/2011 15:37'!
readVisits: visitedParts
	"Make components of the introspectee which are 
	proxies to become real persistents.
	Do this deeply (sub components)."
	
	introspectee class isMeta ifTrue:[^introspectee realPersistent].
	introspectee deservesPersistentRead ifFalse:[^introspectee realPersistent].
	(visitedParts includes: introspectee) ifTrue:[^introspectee realPersistent].
	visitedParts add: introspectee.
	
	introspectee do:[:anElement|
		anElement deservesPersistentRead ifTrue: [
			anElement subPersistentReadVisits: visitedParts].
		anElement isODBReference ifTrue:[
			introspectee 
				replaceAll: anElement 
				with: anElement realPersistent]].

	^ introspectee realPersistent! !

!ARPersistentReader methodsFor: 'actions' stamp: 'sebastiansastre 9/24/2011 13:58'!
read
	"Make components of the introspectee which are 
	proxies to become real persistents.
	Do this deeply (sub components)."

	^ self readVisits: IdentitySet new! !

!ARPersistentReader methodsFor: 'actions' stamp: 'sebastiansastre 9/30/2011 14:47'!
readVisits: visitedParts
	"Make components of the introspectee which are 
	proxies to become real persistents.
	Do this deeply (sub components)."

	introspectee class isMeta ifTrue:[^introspectee realPersistent].
	introspectee deservesPersistentRead ifFalse:[^introspectee realPersistent].
	(visitedParts includes: introspectee) ifTrue:[^introspectee realPersistent].
	visitedParts add: introspectee.
	
	introspectee class allInstVarNames do:[:name| |var| 
		var := introspectee instVarNamed: name.
		var deservesPersistentRead ifTrue:[
			var subPersistentReadVisits: visitedParts].
		var isODBReference ifTrue:[
			introspectee instVarNamed: name put: var realPersistent]].

	^ introspectee realPersistent! !

ARPersistentIntrospector subclass: #ARPersistentWriter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Introspection'!

Object subclass: #ARQueryResults
	instanceVariableNames: 'name results'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Core'!

!ARQueryResults class methodsFor: 'public' stamp: 'sebastian.sastre 11/25/2009 16:53'!
for: someActiveRecords named: aString

	^ self new
		name: aString;
		results: someActiveRecords;
		yourself! !

!ARQueryResults methodsFor: 'accessing' stamp: 'sebastian.sastre 11/25/2009 16:46'!
name
	^ name! !

!ARQueryResults methodsFor: 'accessing' stamp: 'sebastian.sastre 11/25/2009 16:46'!
name: anObject
	name := anObject! !

!ARQueryResults methodsFor: 'actions' stamp: 'sebastian.sastre 11/25/2009 16:47'!
purge 
	"Purifies this cache slot.
	Note: this may help a bit the work of the garbage collector"
	
	results := nil! !

!ARQueryResults methodsFor: 'accessing' stamp: 'sebastian.sastre 11/25/2009 16:46'!
results
	^ results! !

!ARQueryResults methodsFor: 'accessing' stamp: 'sebastian.sastre 11/25/2009 16:46'!
results: anObject
	results := anObject! !

Object subclass: #ARStorage
	instanceVariableNames: 'containerPolicy validations'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Core'!

!ARStorage commentStamp: 'sebastian.sastre 11/19/2009 20:29' prior: 0!
An ARStorage is the guy that actually adds, removes and updates aggregates in the repository so everybody can be confident.
It does the validations before saving and it can destroy dependents (it goes deeply if it has to).
To add aggregates it follows the defined container chosing policy (expected to be invariant on the storage life-span).
!

!ARStorage class methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 20:42'!
defaultContainerPolicy
	"Answers the policy to chose container used by default"
	
	"OmniBase has a default container and this policy will 
	decide to send ot it every new persistent object."
	^ ARDefaultContainer ! !

!ARStorage class methodsFor: 'instance creation' stamp: 'sebastian.sastre 11/19/2009 20:39'!
new

	^ super new
		initialize;
		yourself! !

!ARStorage methodsFor: 'actions' stamp: 'sebastiansastre 10/17/2011 21:45'!
add: anAggregate
	"Adds anAggregate to the repository
	Note: there are validations and policies to meet 
	the application's model required data consistence"
	| containerName attributes |
	
	containerName := containerPolicy containerNameFor: anAggregate in: self.

	ARCurrentContainer use: containerName during:[|aggregateIndices|
		self makeAggregatePersistent: anAggregate in: containerName.

		attributes := self indexedAttributesFor: anAggregate.
		aggregateIndices := self indicesFor: anAggregate class.
		aggregateIndices ifNil:[^ ARNoIndicesForAggregate signal].
		aggregateIndices do: [:index | 
			index add: anAggregate attributes: attributes for: self]]! !

!ARStorage methodsFor: 'actions' stamp: 'sebatian.sastre 4/28/2010 22:47'!
addIndex: anIndex
	"Adds anIndex to this storage.
	It'll be installed in the repository"
	| containerName |

	(self isIndexing: anIndex aggregateClass with: anIndex selector) ifTrue:[
		self error: 'Hey, that index is already in the store.'].
	
	containerName := containerPolicy containerNameForIndex: anIndex in: self.
	
	ARCurrentContainer use: containerName during:[
		self currentTransaction 
			makePersistent: anIndex 
			in: containerName.

		(self indices 
			at: anIndex aggregateClassName 
			ifAbsentPut:[
				self currentTransaction 
					makePersistent: OrderedCollection new
					in: containerName]) add: anIndex; markDirty.
		anIndex resetIndexingSupport.
		self currentTransaction 
			makePersistent: anIndex indexingSupport
			in: containerName.
		self markDirty]! !

!ARStorage methodsFor: 'private' stamp: 'sebastian.sastre 11/19/2009 22:19'!
containerFor: anAggregate
	"Answers the container to be used for anAggregate"
	
	^ containerPolicy containerFor: anAggregate in: self! !

!ARStorage methodsFor: 'private' stamp: 'sebastian.sastre 11/19/2009 22:17'!
containerNameFor: anAggregate
	"Answers the name of the container to be used for anAggregate"
	
	^ containerPolicy containerNameFor: anAggregate in: self! !

!ARStorage methodsFor: 'private' stamp: 'sebastian.sastre 11/19/2009 23:28'!
containerNameForAttributeIndex: anIndex
	"Answers the name of the container to be used for anAggregate"
	
	^ containerPolicy containerNameForAttributeIndex: anIndex in: self! !

!ARStorage methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 20:39'!
containerPolicy
	^ containerPolicy! !

!ARStorage methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 20:39'!
containerPolicy: anObject
	containerPolicy := anObject! !

!ARStorage methodsFor: 'private' stamp: 'sebastian.sastre 11/19/2009 23:20'!
currentTransaction

	^ OmniBase currentTransaction! !

!ARStorage methodsFor: 'queries' stamp: 'sebastian.sastre 5/11/2010 22:14'!
find: anAggregateClass from: start to: end of: aSelector
	"Answers the aggregates of anAggregateClass
	indexed by aSelector that are from 
	the occurrence number start 
	up to the occurrence number end"

	^ (self indexFor: anAggregateClass selector: aSelector) from: start to: end for: self! !

!ARStorage methodsFor: 'queries' stamp: 'sebastian.sastre 4/24/2010 14:02'!
find: anAggregateClass where: aSelectorSymbol from: start to: end
	"Answers the instance of anAggregateClass who's
	aSelectorSymbol answers aValue that is included in the given interval. 
	Or.. nil if not found."
	
	| index result |
	index := self indexFor: anAggregateClass selector: aSelectorSymbol.
	index hasSortSelector
		ifTrue: [result := TreeSet sortSelector: index sortSelector equalitySelector: #=]
		ifFalse: [result := TreeSet sortSelector: aSelectorSymbol equalitySelector: #isIdenticalTo:].
	
	index
		from: start 
		to: end 
		do: [:anAggregate| result add: anAggregate]
		for: self.
		
	^ result! !

!ARStorage methodsFor: 'queries' stamp: 'sebastian.sastre 4/24/2010 12:03'!
find: anAggregateClass where: aSelectorSymbol includes: aValue
	"Answers the instances of anAggregateClass who's
	aSelectorSymbol includes aValue (among many others). 
	Or.. nil if not found."

	^ (self indexFor: anAggregateClass selector: aSelectorSymbol) includes: aValue for: self! !

!ARStorage methodsFor: 'queries' stamp: 'sebastian.sastre 12/23/2010 07:57'!
find: anAggregateClass where: aSelectorSymbol is: aValue
	"Answers the instance of anAggregateClass who's
	aSelectorSymbol answers aValue. 
	Or.. nil if not found."

	^ (self indexFor: anAggregateClass selector: aSelectorSymbol) at: aValue for: self! !

!ARStorage methodsFor: 'testing' stamp: 'sebastian.sastre 4/7/2010 17:21'!
hasIndicesFor: anAggregateClass

	^ (self indicesFor: anAggregateClass) notNilNorEmpty! !

!ARStorage methodsFor: 'testing' stamp: 'sebastian.sastre 11/21/2009 15:32'!
hasUsableIndices
	"Answers true if all the indices in the receiver are usable"
	^ self indices values allSatisfy:[:anIndex| 
		self isUsableIndex: anIndex]! !

!ARStorage methodsFor: 'accessing' stamp: 'sebastiansastre 6/16/2011 16:29'!
indexFor: anAggregateClass selector: aSelectorSymbol
	"Answers the index that is the maintainer
	of anAggregateClass kind of aggregates on aSelectorSymbol values"

	^ self 
		indexFor: anAggregateClass
		selector: aSelectorSymbol 
		ifAbsent:[ARNoIndexForAttribute signal] ! !

!ARStorage methodsFor: 'accessing' stamp: 'sebastiansastre 6/16/2011 16:29'!
indexFor: anAggregateClass selector: aSelectorSymbol ifAbsent: aBlock
	"Answers the index that is the maintainer
	of anAggregateClass kind of aggregates on aSelectorSymbol values"

	^ [:indices|
		indices ifNil:[ARNoIndexForAggregate signal].
		indices 
			detect: [:anIndex| anIndex selector = aSelectorSymbol]
			ifNone: aBlock] value: (self indicesFor: anAggregateClass)
! !

!ARStorage methodsFor: 'private' stamp: 'sebastian.sastre 11/21/2009 15:28'!
indexOfIdexedAttributesFor: anAggregate
	"Answers the (potentially big) dictionary 
	of indexed attributes for anAggregate kind of aggregates.
	Note: this indirection is to have one per aggregate.
	It'll be lazily added when necessary"

	^ OmniBase root 
		at: (self indexedAttributesKeyFor: anAggregate) 
		ifAbsentPut: [OmniBase newBTreeIdentityDictionary makePersistent]
		"ifAbsentPut: [ARBTree order: 32]"
! !

!ARStorage methodsFor: 'accessing' stamp: 'sebastiansastre 12/4/2011 21:20'!
indexedAttributesFor: anAggregate
	"Answers the dictionary with the values related to anAggregate.
	Note: the dictionary stores associations of attribute -> value
	where attribute is the selector of the aggregate for a given value"
	
	| containerName |

	containerName := containerPolicy indexedAttributesContainerNameFor: anAggregate in: self.
	
	^ ARCurrentContainer use: containerName during:[| indexedAttributesIndex |
		indexedAttributesIndex := self indexOfIdexedAttributesFor: anAggregate.
		indexedAttributesIndex 
			at: anAggregate id
			ifAbsent: [
				indexedAttributesIndex
					at: anAggregate id
					put: (self currentTransaction
							makePersistent: Dictionary new
							in: containerName)]]	
	! !

!ARStorage methodsFor: 'private' stamp: 'sebastian.sastre 11/20/2009 21:50'!
indexedAttributesKeyFor: anAggregate
	"Answers the key under which the indexed 
	attributes are to be found."
	
	^ anAggregate class name, '-indexedAttributes'! !

!ARStorage methodsFor: 'accessing' stamp: 'sebastian.sastre 11/21/2009 15:32'!
indices
	^ self root at: #indices! !

!ARStorage methodsFor: 'accessing' stamp: 'sebastian.sastre 11/21/2009 15:31'!
indicesFor: anAggregateClass
	"Answers the indices that are maintainers 
	of anAggregateClass kind of aggregates"
	
	^ self indices at: anAggregateClass name ! !

!ARStorage methodsFor: 'initializing' stamp: 'sebastian.sastre 11/21/2009 16:47'!
initialize

	super initialize.
	
	containerPolicy := self class defaultContainerPolicy	new.
	self root at: #indices put: (OmniBase newBTreeDictionary: 32) makePersistent.
	self makePersistent ! !

!ARStorage methodsFor: 'testing' stamp: 'sebastian.sastre 11/26/2009 17:14'!
isIndexing: aClass with: aSelector
	"Answers true if the receiver is indexing aClass 
	with an accessor invoked by aSelector"

	^ self indices values anySatisfy: [:indices| 
		indices anySatisfy:[:index|
			(index indexesClass: aClass) and:[
			index selector = aSelector]]]! !

!ARStorage methodsFor: 'testing' stamp: 'sebastian.sastre 11/19/2009 23:13'!
isUsableIndex: anIndex
	"Answers true if anIndex is usable in this system (image)."
	
	^ Smalltalk includesKey: anIndex aggregateClassName! !

!ARStorage methodsFor: 'actions' stamp: 'sebastiansastre 9/30/2011 12:46'!
makeAggregatePersistent: anAggregate in: aContainerName
	
	self currentTransaction makePersistent: anAggregate in: aContainerName.

	"The checkpoint here is to let the transaction actually write the aggregate so it gets the
	real objectID at the container."
	self currentTransaction checkpoint.
	anAggregate setOid; markDirty
! !

!ARStorage methodsFor: 'actions' stamp: 'sebastian.sastre 9/27/2010 16:21'!
remove: anAggregate
	"Remove anAggregate from the repository
	Note: there are validations and policies to meet 
	the application's model required data consistence"
	| containerName attributes |
	
	containerName := containerPolicy containerNameFor: anAggregate in: self.

	ARCurrentContainer use: containerName during:[
		attributes := self indexedAttributesFor: anAggregate.
		(self indicesFor: anAggregate class) do: [:index | 
			index remove: anAggregate attributes: attributes for: self].

		(self indexedAttributesFor: anAggregate) 
			removeKey: anAggregate 
			ifAbsent: []]! !

!ARStorage methodsFor: 'actions' stamp: 'sebastian.sastre 4/24/2010 17:51'!
removeAllIndicesOf: anAggregateClass
	"Removes all the indices dedicated to anAggregateClass in this storage."

	| containerName |
	
	containerName := containerPolicy containerFor: anAggregateClass in: self.
	
	ARCurrentContainer use: containerName during:[
		self indices removeKey: anAggregateClass name; markDirty.
		self markDirty]! !

!ARStorage methodsFor: 'actions' stamp: 'sebastian.sastre 4/24/2010 17:01'!
removeIndex: anIndex
	"Removes anIndex from this storage"
	| containerName |
	
	containerName := containerPolicy containerNameForIndex: anIndex in: self.

	ARCurrentContainer use: containerName during:[
		(self indices 
			at: anIndex aggregateClassName
			ifAbsent:[^nil]) remove: anIndex; markDirty.
		self indices markDirty.
		anIndex removeIndexingSupport.
		self markDirty]! !

!ARStorage methodsFor: 'actions' stamp: 'sebastiansastre 6/16/2011 16:29'!
removeIndexOf: aSelector in: anAggregateClass
	"Removes all the indices dedicated to anAggregateClass in this storage."

	| exIndex |

	exIndex := self indexFor: anAggregateClass selector: aSelector ifAbsent:[^nil].
	self removeIndex: exIndex
	! !

!ARStorage methodsFor: 'private' stamp: 'sebastian.sastre 11/21/2009 15:34'!
root
	
	^ self currentTransaction root! !

!ARStorage methodsFor: 'accessing' stamp: 'sebastian.sastre 11/21/2009 15:30'!
unusableIndices
	"Answers the problematic indices."
	^ self indices reject:[:anIndex| 
		self isUsableIndex: anIndex]! !

!ARStorage methodsFor: 'actions' stamp: 'sebastiansastre 10/17/2011 22:10'!
update: anAggregate
	"Updates the indices for anAggregate in the repository
	Note: there are validations and policies to meet 
	the application's model required data consistence"
	| containerName attributes |

	containerName := containerPolicy containerNameFor: anAggregate in: self.

	ARCurrentContainer use: containerName during:[
		self makeAggregatePersistent: anAggregate in: containerName.
		attributes := self indexedAttributesFor: anAggregate.
		(self indicesFor: anAggregate class) do: [:index | 
			index update: anAggregate attributes: attributes for: self]]! !

!ARStorage methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 20:39'!
validations
	^ validations! !

!ARStorage methodsFor: 'accessing' stamp: 'sebastian.sastre 11/19/2009 20:39'!
validations: anObject
	validations := anObject! !

Object subclass: #ARWorkbench
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Aggregate-Test'!

!ARWorkbench class methodsFor: 'actions' stamp: 'sebastiansastre 3/6/2012 16:29'!
addPeople
	
	self odb commit:[
		1000 timesRepeat:[ARPersonMock testPerson save]].
! !

!ARWorkbench class methodsFor: 'accessing' stamp: 'sebastiansastre 3/6/2012 16:29'!
delayAt

	| guy delay |

	self odb readOnly:[
		guy := ARPersonMock findAll asSet anyOne.
		delay := [1000 timesRepeat:[|g|
					g := ARPersonMock atId: guy id]] timeToRun].

	^ delay! !

!ARWorkbench class methodsFor: 'accessing' stamp: 'sebastiansastre 3/6/2012 16:29'!
delayFreshAt

	| guy delay |

	self odb readOnly:[
		guy := ARPersonMock findAll asSet anyOne.
		delay := [1000 timesRepeat:[|g|
					g := guy class at: guy id]] timeToRun].
	^ delay! !

!ARWorkbench class methodsFor: 'accessing' stamp: 'sebastiansastre 3/6/2012 16:29'!
delayFreshOID

	| guy delay |

	self odb readOnly:[
		guy := ARPersonMock findAll asSet anyOne.
		delay := [1000 timesRepeat:[|g|
					g := guy freshOID]] timeToRun].
	^ delay! !

!ARWorkbench class methodsFor: 'accessing' stamp: 'sebastiansastre 3/6/2012 16:29'!
delayOID

	| pal delay oid |

	self odb readOnly:[
		pal := ARPersonMock findAll asSet anyOne.
		oid := pal odbGetOID.
		delay := [1000 timesRepeat:[|g|
					g := ARPersonMock currentTransaction objectAt: oid]] timeToRun].

	^ delay! !

!ARWorkbench class methodsFor: 'accessing' stamp: 'sebastiansastre 9/4/2011 11:33'!
odb 

	^ Flow helper aggregate workerDatabase.
	
! !

!ARWorkbench class methodsFor: 'actions' stamp: 'sebastiansastre 3/6/2012 16:29'!
removeAllPeople
	
	self odb commit:[
		ARPersonMock findAll do:[:e| e destroy]].
! !

!Object class methodsFor: '*aggregate' stamp: 'sas 10/16/2008 20:15'!
deservesPersistentRead
	"Answers true if the receiver deserves 
	to be read (or sub read) as persistent"
	
	^ true! !

!Object class methodsFor: '*aggregate' stamp: 'sebastiansastre 9/25/2011 02:28'!
persistentReaderSupport

	^ ARPersistentReader! !

!Object methodsFor: '*aggregate' stamp: 'sas 8/26/2008 22:19'!
aggregate
	"Answers the receiver.
	Applicability: useful to make proxies fail into the real persistent"
	^ self ! !

!Object methodsFor: '*aggregate' stamp: 'sebastiansastre 10/5/2011 11:57'!
current
	"Answer the receiver for any object but aggregates.
	All objects are considered as the current (inthe assumed current transation).
	Only aggregates can be re-read from in the current transaction.
	Meant to be polimorphic for aggreggates"
	^ self ! !

!Object methodsFor: '*aggregate' stamp: 'sas 10/16/2008 20:21'!
deservesPersistentRead
	"Answers true if the receiver deserves 
	to be read (or sub read) as persistent"
	
	^ self class deservesPersistentRead! !

!Object methodsFor: '*aggregate' stamp: 'sebastian.sastre 12/11/2010 23:28'!
fresh
	"Answer the receiver for any object but aggregates.
	All objects are considered as fresh as they can be.
	Only aggregates can be re-read fresh from the repository.
	Meant to be polimorphic for aggreggates"
	^ self ! !

!Object methodsFor: '*aggregate' stamp: 'sebastian.sastre 2/16/2010 11:44'!
isAggregate 

	^ false! !

!Object methodsFor: '*aggregate' stamp: 'sebastian.sastre 11/26/2009 10:59'!
isMany
	"Answers true if the receiver holds many objects.
	E.g., it implements #do:"
	^ self respondsTo: #do:! !

!Object methodsFor: '*aggregate' stamp: 'sas 10/16/2008 19:56'!
persistentRead
	"Return the receiver with real persistent instead of proxies."

	^ self persistentReader read! !

!Object methodsFor: '*aggregate' stamp: 'sas 10/16/2008 19:56'!
persistentReader
	"Return a persistent reader for the receiver."

	^ self persistentReaderSupport for: self! !

!Object methodsFor: '*aggregate' stamp: 'sas 10/16/2008 19:57'!
persistentReaderSupport
	"Return a persistent reader support for the receiver."

	^ self class persistentReaderSupport ! !

!Object methodsFor: '*aggregate' stamp: 'sebastian.sastre 11/22/2009 14:03'!
realPersistent
	"Answers the receiver.
	Applicability: useful to make proxies fail into the real persistent object"
	^ self ! !

!Object methodsFor: '*aggregate' stamp: 'sas 10/16/2008 20:11'!
subPersistentReadVisits: visitedParts
	"Return the receiver with real persistent instead of proxies."

	^ self persistentReader readVisits: visitedParts! !
ARODBPool initialize!
ARAggregate initialize!
